import astropy.units as u
import numpy as np
from astropy.coordinates import Angle
from astropy.time import Time
from scipy.optimize import minimize_scalar

from dysh.util.gaincorrection import GBTGainCorrection


class TestGainCorrection:
    """Test the GBT Gain Correction functions"""

    def setup_method(self):
        self.gbtgc = GBTGainCorrection()
        self.angles = Angle(np.arange(91), unit=u.degree)
        self.dates = Time(
            [
                "2001-02-03",
                "2002-02-28",
                "2004-07-07",
                "2008-08-04",
                "2009-11-01",
                "2014-04-30",
                "2014-05-01",
                "2018-06-14",
            ]
        )

    def test_surface_error(self):
        answer = np.array([490.0, 490.0, 440.0, 418.0, 240.0, 240.0, 230.0, 230.0]) * u.micron
        for i in range(len(self.dates)):
            se = self.gbtgc.surface_error(self.dates[i])
            assert se == answer[i]

    def test_gaincorrection_factor(self):
        # first test use symmetric angles around 45
        angles = Angle([0.0, 25.0, 40.0, 45.0, 50, 65.0, 90.0], unit=u.degree)
        # answer = np.array([])
        for i in range(len(self.dates)):
            gc1 = self.gbtgc.gain_correction(angles, self.dates[i], zd=True)
            gc2 = self.gbtgc.gain_correction(angles, self.dates[i], zd=False)
            assert all(gc1 == gc2[::-1])
        # now try to reproduce portion of figure 1 in GBT memo 301
        angles = Angle(np.arange(10, 90, 5), unit=u.degree)
        scale = 0.983  # 2009b gmax from Dave Frayer email 12/18/2024
        gc = scale * self.gbtgc.gain_correction(angles, self.dates[4], zd=False)
        # measured using graphreader.com on a screenshot of Figure 1 from GBT memo 301 for "2009b" curve
        answer = np.array(
            [
                0.777,
                0.819,
                0.858,
                0.893,
                0.922,
                0.944,
                0.961,
                0.974,
                0.981,
                0.985,
                0.981,
                0.971,
                0.957,
                0.939,
                0.912,
                0.882,
            ]
        )
        assert np.abs(np.mean(gc - answer)) < 0.001
        # measured using graphreader.com on a screenshot of Figure 1 from GBT memo 301 for "2009a" curve
        gc *= 0.728  # 2009a gmax from Dave Frayer email 12/18/2024.  Yes it is a double scale 0.983*0.728
        answer = np.array(
            [
                0.565,
                0.597,
                0.625,
                0.65,
                0.67,
                0.688,
                0.701,
                0.71,
                0.715,
                0.716,
                0.714,
                0.706,
                0.696,
                0.683,
                0.665,
                0.642,
            ]
        )
        assert np.abs(np.mean(gc - answer)) < 0.001
        # 2014 curve.
        gc = self.gbtgc.gain_correction(angles, self.dates[-1], zd=False)
        # measured using graphreader.com on a screenshot of Figure 1 from GBT memo 301 for "2014" curve
        # harder to do because of graph crowding.
        answer = np.array(
            [0.987, 0.991, 0.994, 0.997, 0.999, 1.0, 1.0, 1.0, 1.0, 1.0, 0.998, 0.995, 0.991, 0.987, 0.983, 0.978]
        )
        assert np.abs(np.mean(gc - answer)) < 0.001
        #
        # 2003 curve was not generated using G(ZD) so we do not test it here.
        # according to Frayer it is (IDL code):
        # ;;No Gravity corrections Condon 2003 PTCS/PN/31.1
        # ;;max eta_a=43% (390um rms) scaled to max of 60% fac=0.43/0.60=0.717
        # nograv=0.717*exp(-1.0*((xel-52.0)/54.)^2.)

    def test_aperture_efficiency(self):
        """This attempts to reproduce Table 2 in GBT Memo 301"""
        freqs = np.array([10.0, 30.0, 43.0, 80.0, 110.0]) * u.GHz
        answer = np.array(
            [
                [0.69, 0.56, 0.43, 0.13, 0.03],  # 2003
                [0.69, 0.56, 0.43, 0.13, 0.03],  # 2009a
                [0.70, 0.65, 0.59, 0.37, 0.21],  # 2009b
                [0.70, 0.65, 0.60, 0.39, 0.23],  # 2014
            ]
        )
        f = self.gbtgc.gain_correction
        i = 0
        for d in self.dates[[5, 7]]:
            # first find the elevation angle where the gain curve reaches a maximum
            maxpoint = minimize_scalar(
                lambda x: -f(angle=x * u.degree, date=d, zd=False), bounds=[0, 90], method="bounded"
            )
            # Evaluate the aperture efficiency at the given requencies and the elevation of the gain maximum
            a = maxpoint.x * u.degree
            if i < 2:
                # For 2003 and 2009a, a smaller surface error is used in memo 301 than the standard in our gain table.
                # So pass in the memo value in these cases.
                ap = self.gbtgc.aperture_efficiency(freqs, a, d, zd=False, eps0=390 * u.micron)
            else:
                ap = self.gbtgc.aperture_efficiency(freqs, a, d, zd=False)
            apr = np.round(ap, decimals=2)  # this is all the precision the table has
            # print(np.abs(apr - answer[i]))
            assert all(apr == answer[i])
            i += 1

    def test_airmass(self):
        angles = Angle([0.0, 25.0, 40.0, 45.0, 50, 65.0, 90.0], unit=u.degree)
        answer = np.array([37.5541407, 2.35964333, 1.5501963, 1.40793864, 1.29840534, 1.09473653, 0.99060048])
        am = self.gbtgc.airmass(angles, zd=False)
        assert np.mean(am - answer) < 1e-10
        assert all(self.gbtgc.airmass(angles, zd=True) == am[::-1])



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Union

import astropy.units as u
import numpy as np
from astropy.coordinates import Angle
from astropy.table import QTable
from astropy.time import Time
from astropy.units.quantity import Quantity

from ..log import logger
from ..util import get_project_configuration

__all__ = ["BaseGainCorrection", "GBTGainCorrection"]


class BaseGainCorrection(ABC):
    r"""This class is the base class for gain corrections. It is intended to be subclassed for
    specific antennas. Subclasses will be used to calculate various gain corrections to go
    from antenna temperature to other scales like brightness temperature of flux density.
    Subclasses can implement the following attributes:
        * `ap_eff_0`
            long wavelength aperture efficiency (number between 0 and 1), i.e., in the absence of surface errors,
            $\lambda >> \epsilon_0$.

        * `epsilon_0`
            default rms surface accuracy with units of length (`~astropy.units.quanitity.Quantity`)

        * `physical_aperture`
            antenna physical aperture with units of length**2 (`~astropy.units.quanitity.Quantity`)
    """

    def __init__(self):
        self.ap_eff_0 = 1.0
        self.epsilon_0 = 100 * u.micron
        self.physical_aperture = 1.0 * u.m * u.m

    @abstractmethod
    def airmass(self, angle: Union[Angle, Quantity], zd: bool, **kwargs) -> Union[float, np.ndarray]:
        """Computes the airmass at given elevation(s) or zenith distance(s).
        Subclasses should implement an airmass function specific to their application.

        Parameters
        ----------
        angle : `~astropy.coordinates.Angle` or `~astro.units.quantity.Quantity`
            The elevation(s) or zenith distance(s) at which to compute the airmass

        zd: bool
            True if the input value is zenith distance, False if it is elevation.

        **kwargs : Any
            Other possible parameters that affect the airmass, e.g. weather data.

        Returns
        -------
            airmass - float or `~numpy.ndarray`
            The value(s) of the airmass at the given elevation(s)/zenith distance(s)
        """
        pass

    @abstractmethod
    def aperture_efficiency(self, specval: Quantity, **kwargs) -> Union[float, np.ndarray]:
        """
        Calculate the antenna aperture efficiency.

        Parameters
        ----------
        specval : `~astro.units.quantity.Quantity`
            The spectral value -- frequency or wavelength -- at which to compute the efficiency

        **kwargs : Any
            Other possible parameters that affect the aperture efficiency, e.g., elevation angle.

        Returns
        -------
            aperture_efficiency  - float or `~numpy.ndarray`
            The value(s) of the aperture efficiency at the given frequency.
            The return value(s) are float(s) between zero and one.

        """
        pass


class GBTGainCorrection(BaseGainCorrection):
    """Gain correction class and functions specific to the Green Bank Telescope"""

    def __init__(self, gain_correction_table: Path = None):
        """
        Parameters
        -----------
        gain_correction_table : str or `pathlib.Path`
             File to read that contains the parameterized gain correction as a function
             of zenith distance and time
            (see `GBT Memo 301 <https://library.nrao.edu/public/memos/gbt/GBT_301.pdf>`_).
             Must be in an `~astropy.table.QTable` readable format.
             Default None will usedysh's internal GBT gain correction table.
        """
        if gain_correction_table is None:
            gain_correction_table = get_project_configuration() / "gaincorrection.tab"
        self._gct = QTable.read(gain_correction_table, format="ascii.ecsv")
        self._gct.sort("Date")  # just in case it ever is written unsorted.
        self.app_eff_0 = 0.71
        self.epsilon_0 = 230 * u.micron
        self.physical_aperture = 7853.9816 * u.m * u.m

    @property
    def gain_correction_table(self):
        """The table containing the parameterized gain correction as a fucntion of zenith distance and time"""
        return self._gct

    def airmass(self, angle: Union[Angle, Quantity], zd: bool = False, **kwargs) -> Union[float, np.ndarray]:
        """
        Computes the airmass at given elevation(s) or zenith distance(s).  The formula used is

        :math:`A = -0.0234 + 1.014/sin(El+5.18/(El+3.35))`

        for elevation in degrees. This function is specific for the GBT location derived
        from vertical weather data. Source: `(Maddalena 2007) <https://www.gb.nrao.edu/~rmaddale/GBT/Maddalena_HighPrecisionCalibration.pdf>`_

        Parameters
        ----------
        angle :  `~astropy.coordinates.Angle` or `~astro.units.quantity.Quantity`
            The elevation(s) or zenith distance(s) at which to compute the airmass

        zd: bool
            True if the input value is zenith distance, False if it is elevation. Default: False

        Returns
        -------
            airmass - float or `~numpy.ndarray`
            The value(s) of the airmass at the given elevation(s)/zenith distance(s)

        """
        ang_deg = angle.to(u.degree)
        if zd:
            ang_deg = 90.0 * u.degree - ang_deg

        c0 = -0.0234
        c1 = 5.18
        c2 = 3.35
        c3 = 1.014
        d = ang_deg.value + c1 / (ang_deg.value + c2)

        return c0 + c3 / np.sin(np.radians(d))

    def _get_gct_index(self, date: Time) -> int:
        """
        locate the row in GC table that is applicable to the input date.
        Assumes table is sorted (happens in constructor)!

        Parameters
        ----------
        date : `~astropy.time.Time`
            Date of observation

        Returns
        -------
        int
            Index to use from gain correction table

        """
        tablen = len(self._gct)
        index = tablen - 1
        for i in range(tablen):
            if date <= self._gct["Date"][i]:
                index = i
                break
        return index

    def surface_error(self, date: Time) -> Quantity:
        """
        Lookup the applicable surface error in the gain correction table
        for the observation date.

        Parameters
        ----------
        date : `~astropy.time.Time`
            Date of observation

        Returns
        -------
            `~astropy.units.quantity.Quantity`
                Surface error for the given date.

        """
        i = self._get_gct_index(date)
        return self._gct[i]["Surface Error"]

    def gain_correction(
        self,
        angle: Union[Angle, Quantity],
        date: Time,
        zd: bool = True,
    ) -> Union[float, np.ndarray]:
        r"""
                Compute the gain correction scale factor, to be used in the aperture efficiency
                calculation. The factor is a float between zero and 1.
                (See `GBT Memo 301 <https://library.nrao.edu/public/memos/gbt/GBT_301.pdf>`_).
        The factor is
                determined by:

                :math:`G = A0 + A1*ZD + A2*ZD^2`

                where An are the time-dependent coefficients and ZD is the zenith distance angle in degrees.

                Parameters
                ----------
                angle :  `~astropy.coordinates.Angle` or `~astro.units.quantity.Quantity`
                    The elevation(s) or zenith distance(s) at which to compute the gain correction factor

                date  : `~astropy.time.Time`
                    The date at which to cmopute the gain correction factor

                zd: bool
                    True if the input value is zenith distance, False if it is elevation. Default: False

                Returns
                -------
                    gain_correction - float or `~numpy.ndarray`
                    The gain correction scale factor(s) at the given elevation(s)/zenith distance(s)
        """
        i = self._get_gct_index(date)
        a0 = self._gct[i]["A0"]
        a1 = self._gct[i]["A1"]
        a2 = self._gct[i]["A2"]
        logger.debug(f"Using {a0=} {a1=} {a2=} for {date}")
        ang_deg = angle.to(u.degree)
        if not zd:
            z = 90.0 - ang_deg.value
        else:
            z = ang_deg.value

        return a0 + a1 * z + a2 * z * z

    def aperture_efficiency(
        self, specval: Quantity, angle: Union[Angle, Quantity], date: Time, zd=False, eps0=None, **kwargs
    ) -> Union[float, np.ndarray]:
        r"""
        Compute the aperture efficiency, as a float between zero and 1. The aperture
        efficiency :math:`\eta_a`, is determined by:

                :math:`\eta_a = \eta_0 G(ZD) \exp(-(4\pi\epsilon_0/\lambda)^2)`

        where :math:`\eta_0` is the long wavelength aperture efficiency, :math:`G(ZD)` is the gain correction factor
        at a zenith distance :math:`ZD, \epsilon_0` is the surface error, and :math:`\lambda` is the wavelength.

        Parameters
        ----------
        specval : `~astro.units.quantity.Quantity`
            The spectral value -- frequency or wavelength -- at which to compute the efficiency

        angle :  `~astropy.coordinates.Angle` or `~astro.units.quantity.Quantity`
            The elevation(s) or zenith distance(s) at which to compute the efficiency

        date  : `~astropy.time.Time`
            The date at which to cmopute the efficieyncy

        zd: bool
            True if the input value is zenith distance, False if it is elevation. Default: False

        eps0 :  `~astro.units.quantity.Quantity` or None
            The value of :math:`\epsilon_0` to use. If given, must have units of length (typically microns).
            If None, the measured value from observatory testing will be used (See :meth:`surface_error`).

        Returns
        -------
            float or `~numpy.ndarray`
            The aperture efficiency at the given inputs

        """
        coeff = self.app_eff_0 * self.gain_correction(angle, date, zd)
        if eps0 is None:
            eps0 = self.surface_error(date)
        _lambda = specval.to(eps0.unit, equivalencies=u.spectral())
        a = (4.0 * np.pi * eps0 / _lambda) ** 2
        eta_a = coeff * np.exp(-a)  # this will be a Quantity with units u.dimensionless
        return eta_a.value



"""
Core functions for spectral data.
"""

import warnings
from copy import deepcopy
from functools import reduce

import astropy.units as u
import numpy as np
from astropy.convolution import (
    Box1DKernel,
    Gaussian1DKernel,
    Trapezoid1DKernel,
    convolve,
)
from astropy.modeling.fitting import LinearLSQFitter
from astropy.modeling.polynomial import Chebyshev1D, Hermite1D, Legendre1D, Polynomial1D
from scipy import ndimage
from specutils import SpectralRegion
from specutils.fitting import fit_continuum
from specutils.fitting.fitmodels import _strip_units_from_model
from specutils.utils import QuantityModel

from ..coordinates import veltofreq
from ..log import log_function_call, logger
from ..util import minimum_string_match, powerof2


# @todo: allow data to be SpectrumList or array of Spectrum
def average(data, axis=0, weights=None):
    """Average a group of spectra or scans.

    Parameters
    ----------
    data : `~numpy.ndarray`
        The spectral data, typically with shape (nspect,nchan).
    axis : int
        The axis over which to average the data.  Default axis=0 will return the average spectrum
        if shape is (nspect,nchan)
    weights : `~numpy.ndarray`
        The weights to use in averaging.  These might typically be system temperature based.
        The weights array must be the length of the axis over which the average is taken.
        Default: None will use equal weights

    Returns
    -------
    average : `~numpy.ndarray`
        The average along the input axis
    """
    # Spectra that are blanked will have all channels set to NaN
    # indices = ~np.isnan(data)
    # Find indices with ANY spectra (rows) that have a NaN or Inf
    # goodindices = ~np.ma.fix_invalid(data).mask.any(axis=1)
    # c = data[goodindices]
    # if len(c) == 0:
    #    return np.nan
    # Find indices that have any spectra with all channels = NaN
    # badindices = np.where(np.isnan(data).all(axis=1))
    goodindices = find_non_blanks(data)
    if weights is not None:
        return np.average(data[goodindices], axis, weights[goodindices])
    else:
        return np.average(data[goodindices], axis, weights)


def integration_isnan(data):
    """Helper function to calculate a boolean array that indicates whether
    a collection of integrations is blanked.

    Parameters
    ----------
    data : `~numpy.ndarray`
        The spectral data, typically with shape (nspect,nchan).

    Returns
    -------
    blanks : `~numpy.ndarray`
        Array with length nspect with value True where an integration is blanked.
    """

    return np.isnan(data).all(axis=1)


def find_non_blanks(data):
    """
    Finds the indices of integrations that are not blanked.

    Parameters
    ----------
    data : `~numpy.ndarray`
        The spectral data, typically with shape (nspect,nchan).

    Returns
    -------
    blanks : `~numpy.ndarray`
        Array with indices of non-blanked integrations.
    """

    return np.where(~integration_isnan(data))


def find_blanks(data):
    """
    Finds the indices of blanked integrations

    Parameters
    ----------
    data : `~numpy.ndarray`
        The spectral data, typically with shape (nspect,nchan).

    Returns
    -------
    blanks : `~numpy.ndarray`
        Array with indices of blanked integrations.
    """

    return np.where(integration_isnan(data))


def find_nonblank_ints(cycle1, cycle2, cycle3=None, cycle4=None):
    """
    Find the indices of integrations that are not blanked.

    Parameters
    ----------
    cycle1 : `~numpy.ndarray`
        Data for cycle 1. For example, signal with the noise diode off.
    cycle2 : `~numpy.ndarray`
        Data for cycle 2. For example, reference with the noise diode off.
    cycle3 : `~numpy.ndarray`
        Data for cycle 3. For example, signal with the noise diode on.
        Default is `None`.
    cycle4 : `~numpy.ndarray`
        Data for cycle 4. For example, reference with the noise diode on.
        Default is `None`.

    Returns
    -------
    goodrows : `~numpy.array`
        Indices of the non-blanked rows.
    """

    nb1 = find_non_blanks(cycle1)
    nb2 = find_non_blanks(cycle2)
    if cycle3 is not None:
        nb3 = find_non_blanks(cycle3)
    else:
        nb3 = nb1
    if cycle4 is not None:
        nb4 = find_non_blanks(cycle4)
    else:
        nb4 = nb2
    goodrows = reduce(np.intersect1d, (nb1, nb2, nb3, nb4))

    if len(goodrows) != len(cycle1):
        nblanks = len(cycle1) - len(goodrows)
        logger.info(f"Ignoring {nblanks} blanked integration(s).")

    return goodrows


def sort_spectral_region(spectral_region):
    """
    Sort the elements of a `~specutils.SpectralRegion`.

    Parameters
    ----------
    spectral_region : `~specutils.SpectralRegion`
        `~specutils.SpectralRegion` to be sorted.

    Returns
    -------
    sorted_spectral_region : `~specutils.SpectralRegion`
        Sorted `~specutils.SpectralRegion`.
    """

    unit = spectral_region.lower.unit
    bound_list = np.sort([srb.value for sr in spectral_region.subregions for srb in sr]) * unit
    it = iter(bound_list)
    sorted_spectral_region = SpectralRegion(list(zip(it, it)))

    return sorted_spectral_region


def include_to_exclude_spectral_region(include, refspec):
    """
    Convert an inclusion region to an exclude region.

    Parameters
    ----------
    include : list of 2-tuples of int or `~astropy.units.Quantity`, or `~specutils.SpectralRegion`
        List of region(s) to exclude from the fit.  The tuple(s) represent a range in the form [lower,upper], inclusive.
        Examples:

        One channel-based region:

        >>> [11,51]

        Two channel-based regions:

        >>> [(11,51),(99,123)]

        One `~astropy.units.Quantity` region:

        >>> [110.198*u.GHz,110.204*u.GHz].

        One compound `~specutils.SpectralRegion`:

        >>> SpectralRegion([(110.198*u.GHz,110.204*u.GHz),(110.196*u.GHz,110.197*u.GHz)]).

    refspec: `~spectra.spectrum.Spectrum`
        The reference spectrum whose spectral axis will be used
        when converting between `include` and axis units (e.g. channels to GHz).

    Returns
    -------
    exclude_region : `~specutils.SpectralRegion`
        `include` as a region to be excluded.
    """

    unit = "cm"  # This has to be wavelength units. cm are familiar to observers?

    spectral_region = exclude_to_spectral_region(include, refspec)
    spectral_region = spectral_region_to_unit(spectral_region, refspec, unit=unit)
    # Convert unit of reference spectrum to wavelength.
    # For some reason `SpectralRegion.invert_from_spectrum`
    # only works in wavelength units.
    org_unit = refspec._spectral_axis.unit
    refspec._spectral_axis = refspec._spectral_axis.to(unit, equivalencies=refspec.equivalencies)
    exclude_region = spectral_region.invert_from_spectrum(refspec)
    refspec._spectral_axis = refspec._spectral_axis.to(org_unit, equivalencies=refspec.equivalencies)
    exclude_region = spectral_region_to_unit(exclude_region, refspec)
    # Sort the output.
    # For some reason it is scrambled.
    exclude_region = sort_spectral_region(exclude_region)

    return exclude_region


def exclude_to_spectral_region(exclude, refspec, fix_exclude=True):
    """Convert `exclude` to a `~specutils.SpectralRegion`.

    Parameters
    ----------
    exclude : list of 2-tuples of int or `~astropy.units.Quantity`, or `~specutils.SpectralRegion`
        List of region(s) to exclude. The tuple(s) represent a range in the form [lower,upper], inclusive.
        Examples:

        One channel-based region:

        >>> [11,51]

        Two channel-based regions:

        >>> [(11,51),(99,123)]

        One `~astropy.units.Quantity` region:

        >>> [110.198*u.GHz,110.204*u.GHz].

        One compound `~specutils.SpectralRegion`:

        >>> SpectralRegion([(110.198*u.GHz,110.204*u.GHz),(110.196*u.GHz,110.197*u.GHz)]).

    refspec: `~spectra.spectrum.Spectrum`
        The reference spectrum whose spectral axis will be used
        when converting between `exclude` and axis units (e.g. channels to GHz).
    fix_exclude: bool
        If True, fix exclude regions that are out of bounds of the spectral axis to be within the spectral axis.
        Default: True

    Returns
    -------
    sr : `~specutils.SpectralRegion`
        A `~specutils.SpectralRegion` corresponding to `exclude`.
    """

    p = refspec
    sa = refspec.spectral_axis
    lastchan = len(sa) - 1

    o = 1
    # If the spectral axis is inverted, flip the order of the elements.
    if refspec.spectral_axis_direction == "decreasing":
        o = -1

    if exclude is not None:
        # A single SpectralRegion was given.
        if isinstance(exclude, SpectralRegion):
            sr = exclude
        # `list` of `int` or `Quantity` or `SpectralRegion` was given.
        else:
            # If user provided a single list, we have to
            # turn it into a list of tuples. If SpectralRegion
            # took a list argument, we wouldn't have to do this.
            if type(exclude[0]) is not tuple:
                it = iter(exclude)
                exclude = list(zip(it, it))
            try:
                sr = SpectralRegion(exclude)
                # The above will error if the elements are not quantities.
                # In that case use the spectral axis to define the exclusion regions.
            except ValueError:
                # If the user requested to fix the exclude range.
                if fix_exclude:
                    exclude = np.array(exclude)
                    mask = exclude > len(sa)
                    if mask.sum() > 0:
                        msg = f"Setting upper limit to {lastchan}."
                        exclude[exclude > len(sa)] = lastchan
                        warnings.warn(msg)
                # If the spectral_axis is decreasing, flip it.
                sr = SpectralRegion(sa[exclude][:, ::o])

        return sr


def spectral_region_to_unit(spectral_region, refspec, unit=None):
    """
    Change the unit of `spectral_region` to `unit` using the equivalencies of `refspec`.
    If no `unit` is provided, it will change to the units of `refspec._spectral_axis`.

    Parameters
    ----------
    spectral_region : `~specutils.SpectralRegion`
        `~specutils.SpectralRegion` whose units will be converted.
    refspec : `~spectra.spectrum.Spectrum`
        The reference spectrum whose spectral axis will be used
        when converting to `unit` (e.g. channels to GHz).
    unit : str or `~astropy.units.Quantity`
        The target units for `spectral_region`.

    Returns
    -------
    spectral_region : `~specutils.SpectralRegion`
        SpectralRegion with units of `unit`.
    """

    qt = spectral_region.as_table()

    if unit is None:
        unit = refspec._spectral_axis.unit

    lb = qt["lower_bound"].to(unit, equivalencies=refspec.equivalencies)
    ub = qt["upper_bound"].to(unit, equivalencies=refspec.equivalencies)

    return SpectralRegion(list(zip(lb, ub)))


def spectral_region_to_list(spectral_region):
    """
    Turn `spectral_region` into a list of `~specutils.SpectralRegion`.
    Each subregion in `spectral_region` will be a list element.

    Parameters
    ----------
    spectral_region : `~specutils.SpectralRegion`
        `~specutils.SpectralRegion` to convert into a list of `~specutils.SpectralRegion`.

    Returns
    -------
    region_list : list of `~specutils.SpectralRegion`
        Subregions of `spectral_region` in a list of `~specutils.SpectralRegion`.
    """

    region_list = []

    # The continuum fitting routines use a list of `SpectralRegion` as input.
    for r in spectral_region.subregions:
        region_list.append(SpectralRegion([r]))

    return region_list


def region_to_axis_indices(region, refspec):
    """
    Parameters
    ----------
    region : `~specutils.SpectralRegion`
    refspec: `~spectra.spectrum.Spectrum`
        The reference spectrum whose spectral axis will be used
        when converting between exclude and axis units (e.g., channels to GHz).

    Returns
    -------
    indices : 2-tuple of int
        The array indices in `refspec` corresponding to `region.bounds`
    """
    # Spectral region to indices in an input spectral axis.
    # @todo needs to work for multiple spectral regions? or just loop outside this call
    sa = refspec.spectral_axis
    if region.lower.unit != sa.unit:
        # @todo if they are conformable, then allow it and convert
        raise Exception(f"Axis units of region [{region.lower.unit}] and refspec [{sa.unit}] not identical")
    b = [x.value for x in region.bounds]
    indices = np.abs(np.subtract.outer(sa.value, b)).argmin(0)
    return indices


def exclude_to_mask(exclude, refspec):
    # set a mask based on an exclude region
    # mask ~ exclude_to_indices(exclude_to_region())
    pass


def exclude_to_region_list(exclude, spectrum, fix_exclude=True):
    """
    Convert an exclusion region, `exclude`, to a list of `~specutils.SpectralRegion`.
    This is used for baseline fitting.

    Parameters
    ----------
    exclude : list of 2-tuples of int or `~astropy.units.Quantity`, or `~specutils.SpectralRegion`
        List of region(s) to exclude. The tuple(s) represent a range in the form [lower,upper], inclusive.
        Examples:

        One channel-based region:

        >>> [11,51]

        Two channel-based regions:

        >>> [(11,51),(99,123)]

        One `~astropy.units.Quantity` region:

        >>> [110.198*u.GHz,110.204*u.GHz].

        One compound `~specutils.SpectralRegion`:

        >>> SpectralRegion([(110.198*u.GHz,110.204*u.GHz),(110.196*u.GHz,110.197*u.GHz)]).

    spectrum : `~spectra.spectrum.Spectrum`
        The reference spectrum whose spectral axis will be used
        when converting between `exclude` and axis units (e.g. channels to GHz).
    fix_exclude : bool
        See `~spectra.core.exclude_to_spectral_region` for details.
        Default: True

    Returns
    -------
    region_list : list of `~specutils.SpectralRegion`
        Regions defined in `exclude` as a list of `~specutils.SpectralRegion`.
    """

    spectral_region = exclude_to_spectral_region(exclude, spectrum, fix_exclude=fix_exclude)
    spectral_region = spectral_region_to_unit(spectral_region, spectrum)
    region_list = spectral_region_to_list(spectral_region)

    return region_list


@log_function_call()
def baseline(spectrum, order, exclude=None, exclude_region_upper_bounds=True, **kwargs):
    """Fit a baseline to `spectrum`.
    The code uses `~astropy.modeling.fitting.Fitter` and `~astropy.modeling.polynomial` to compute the baseline.
    See the documentation for those modules for details.

    Parameters
    ----------
    spectrum : `~spectra.spectrum.Spectrum`
        The input spectrum.
    order : int
        The order of the polynomial series, a.k.a. baseline order.
    exclude : list of 2-tuples of int or `~astropy.units.Quantity`, or `~specutils.SpectralRegion`
        List of region(s) to exclude from the fit. The tuple(s) represent a range in the form [lower,upper], inclusive.
        Examples:

        One channel-based region:

        >>> [11,51]

        Two channel-based regions:

        >>> [(11,51),(99,123)]

        One `~astropy.units.Quantity` region:

        >>> [110.198*u.GHz,110.204*u.GHz].

        One compound `~specutils.SpectralRegion`:

        >>> SpectralRegion([(110.198*u.GHz,110.204*u.GHz),(110.196*u.GHz,110.197*u.GHz)]).

        Default: no exclude region

    model : str
        One of 'polynomial' or 'chebyshev', Default: 'polynomial'
    fitter : `~astropy.modeling.fitting.Fitter`
        The fitter to use. Default: `~astropy.modeling.fitter.LinearLSQFitter` (with `calc_uncertaintes=True`).
        Be careful when choosing a different fitter to be sure it is optimized for this problem.
    exclude_region_upper_bounds : bool
        Makes the upper bound of any excision region(s) inclusive.
        Allows excising channel 0 for lower-sideband data, and the last channel for upper-sideband data.

    Returns
    -------
    model : `~specutils.utils.QuantityModel`
        Best fit model.
        See `~specutils.fitting.fit_continuum`.
    """
    kwargs_opts = {
        #'show': False,
        "model": "chebyshev",
        "fitter": LinearLSQFitter(calc_uncertainties=True),
        "fix_exclude": True,
        "exclude_action": "replace",  # {'replace','append', None}
    }
    kwargs_opts.update(kwargs)

    available_models = {
        "chebyshev": Chebyshev1D,
        "hermite": Hermite1D,
        "legendre": Legendre1D,
        "polynomial": Polynomial1D,
    }
    model = minimum_string_match(kwargs_opts["model"], list(available_models.keys()))
    if model == None:
        raise ValueError(f'Unrecognized input model {kwargs["model"]}. Must be one of {list(available_models.keys())}')
    sa_min = spectrum.spectral_axis.min().value
    sa_max = spectrum.spectral_axis.max().value
    selected_model = available_models[model](degree=order, domain=(sa_max, sa_min))

    _valid_exclude_actions = ["replace", "append", None]
    if kwargs_opts["exclude_action"] not in _valid_exclude_actions:
        raise ValueError(
            f'Unrecognized exclude region action {kwargs["exclude_region"]}. Must be one of {_valid_exclude_actions}'
        )
    fitter = kwargs_opts["fitter"]
    # print(f"MODEL {model} FITTER {fitter}")
    p = spectrum
    if np.isnan(p.data).all():
        # @todo handle masks
        return None  # or raise exception
    if exclude is not None:
        regionlist = exclude_to_region_list(exclude, spectrum, fix_exclude=kwargs_opts["fix_exclude"])
        if kwargs_opts["exclude_action"] == "replace":
            p._exclude_regions = regionlist
        elif kwargs_opts["exclude_action"] == "append":
            p._exclude_regions.extend(regionlist)
            regionlist = p._exclude_regions
    else:
        # use the spectrum's preset exclude regions if they
        # exist (they will be a list of SpectralRegions or None)
        regionlist = p._exclude_regions

    logger.info(f"EXCLUDING {regionlist}")

    fitted_model = fit_continuum(
        spectrum=p,
        model=selected_model,
        fitter=fitter,
        exclude_regions=regionlist,
        exclude_region_upper_bounds=exclude_region_upper_bounds,
    )

    # astropy's Polynomial1D model does not work well when using
    # a domain other than (-1,1), so we need to remove the units
    # of the model. We use `specutils` methods to do so. This has the
    # downside that the uncertainties of the model are removed.
    if not isinstance(fitted_model, QuantityModel):
        strip_model, _, _ = _strip_units_from_model(fitted_model, p)
        strip_model.domain = fitted_model.domain
        strip_model.window = fitted_model.window
        fitted_model = QuantityModel(strip_model, p.spectral_axis.unit, p.flux.unit)

    return fitted_model


def mean_tsys(calon, caloff, tcal, mode=0, fedge=0.1, nedge=None):
    """
    Get the system temperature from the neighboring calon and caloff, which reflect the state of the noise diode.
    We define an extra way to set the edge size, nedge, if you prefer to use
    number of edge channels instead of the inverse fraction.  This implementation recreates GBTIDL's `dcmeantsys`.

    Parameters
    ----------
        calon : `~numpy.ndarray`-like
            ON calibration

        caloff  :  `~numpy.ndarray`-like
            OFF calibration

        tcal  :  `~numpy.ndarray`-like
            calibration temperature

        mode : int
            mode=0  Do the mean before the division
            mode=1  Do the mean after the division

        fedge : float
            Fraction of edge channels to exclude at each end, a number between 0 and 1. Default: 0.1, meaning the central 80% bandwidth is used

        nedge : int
            Number of edge channels to exclude. Default: None, meaning use `fedge`

    Returns
    -------
        meanTsys : `~numpy.ndarray`-like
            The mean system temperature
    """
    # @todo Pedro thinks about a version that takes a spectrum with multiple SpectralRegions to exclude.
    nchan = len(calon)
    if nedge is None:
        nedge = int(nchan * fedge)
    # Python uses exclusive array ranges while GBTIDL uses inclusive ones.
    # Therefore we have to add a channel to the upper edge of the range
    # below in order to reproduce exactly what GBTIDL gets for Tsys.
    # See github issue #28.
    # Define the channel range once.
    chrng = slice(nedge, -(nedge - 1), 1)

    # Make them doubles. Probably not worth it.
    caloff = caloff.astype("d")
    calon = calon.astype("d")

    if mode == 0:  # mode = 0 matches GBTIDL output for Tsys values
        meanoff = np.nanmean(caloff[chrng])
        meandiff = np.nanmean(calon[chrng] - caloff[chrng])
        meanTsys = meanoff / meandiff * tcal + tcal / 2.0
    else:
        meanTsys = np.nanmean(caloff[chrng] / (calon[chrng] - caloff[chrng]))
        meanTsys = meanTsys * tcal + tcal / 2.0

    # meandiff can sometimes be negative, which makes Tsys negative!
    # GBTIDL also takes abs(Tsys) because it does sqrt(Tsys^2)
    return np.abs(meanTsys)


def sq_weighted_avg(a, axis=0, weights=None):
    # @todo make a generic moment or use scipy.stats.moment
    r"""Compute the mean square weighted average of an array (2nd moment).

    :math:`v = \sqrt{\frac{\sum_i{w_i~a_i^{2}}}{\sum_i{w_i}}}`

    Parameters
    ----------
    a : `~numpy.ndarray`
        The data to average.
    axis : int
        The axis over which to average the data.  Default: 0
    weights : `~numpy.ndarray` or None
        The weights to use in averaging.  The weights array must be the
        length of the axis over which the average is taken.  Default:
        `None` will use equal weights.

    Returns
    -------
    average : `~numpy.ndarray`
        The square root of the squared average of a along the input axis.
    """
    # if weights is None:
    #    w = np.ones_like(a)
    # else:
    #    w = weights
    a2 = a * a
    v = np.sqrt(np.average(a2, axis=axis, weights=weights))
    return v


def tsys_weight(exposure, delta_freq, tsys):
    r"""Compute the system temperature based weight(s) using the expression:
        :math:`w = t_{exp} \times \delta_\nu / T_{sys}^2,`

    If `exposure`, `delta_freq`, or `tsys` parameters are given as `~astropy.units.Quantity`,
    they will be converted to seconds, Hz, K, respectively for the calculation.
    (Not that this really matters since weights are relative to each other)


     >>> import astropy.units as u
     >>> [3*u.s, 4*u.s]   ### WRONG

    Rather, if using `~astropy.units.Quantity`, they have to be `~astropy.units.Quantity` objects, e.g.,

     >>> [3, 4]*u.s ### RIGHT!

    Parameters
    ----------
         exposure : `~numpy.ndarray`, float, or `~astropy.units.Quantity`
             The exposure time, typically given in seconds
         delta_freq : `~numpy.ndarray`, float, or `~astropy.units.Quantity`
             The channel width in frequency units
         tsys : `~numpy.ndarray`, float, or `~astropy.units.Quantity`
             The system temperature, typically in K

    Returns
    -------
         weight : `~numpy.ndarray`
             The weights array
    """

    # Quantitys work with abs and power!
    # Using `numpy.power` like this results in increased
    # precision over the calculation used by GBTIDL:
    # weight = abs(delta_freq) * exposure / tsys**2.
    weight = abs(delta_freq) * exposure * np.power(tsys, -2.0)
    if type(weight) == u.Quantity:
        return weight.value.astype(np.float64)
    else:
        return weight.astype(np.float64)


def get_spectral_equivalency(restfreq, velocity_convention):
    # Yeesh, the doppler_convention parameter for SpectralAxis.to does not match the doppler_convention list for Spectrum1D!
    # This is actually bug in Spectrum1D documentation https://github.com/astropy/specutils/issues/1067
    if "radio" in velocity_convention:
        return u.doppler_radio(restfreq)
    elif "optical" in velocity_convention:
        return u.doppler_optical(restfreq)
    elif "relativistic" in velocity_convention:
        return u.doppler_relativistic(restfreq)
    elif "redshift" in velocity_convention:
        return u.doppler_redshift()
    else:
        raise ValueError(f"Unrecognized velocity convention {velocity_convention}")


def fft_pad(y):
    """
    Pad signal `y` to the next power of 2 using its edge values.

    Parameters
    ----------
    y : 1D `~numpy.ndarray`
        Signal to be padded.

    Returns
    -------
    padded : 1D `~numpy.ndarray`
        Padded signal.
    nskip : int
        Number of samples added to each end of the padded signal.
    """

    nch = len(y)
    pow2 = powerof2(nch)
    pow2 += 1
    newsize = 2**pow2

    padded = np.empty(newsize, dtype=y.dtype)
    npad = newsize - nch
    nskip = npad // 2
    padded[nskip : nskip + nch] = y
    padded[0:nskip] = padded[nskip]
    padded[nskip + nch :] = padded[nskip + nch]

    return padded, nskip


def fft_shift(
    y,
    shift,
    pad=True,
    window=True,
    nan_treatment="fill",
    fill_value=0,
    keep_nan=True,
):
    """
    Shift a signal `y` by `shift` amount using a phase shift.
    This requires taking the inverse FFT of the signal, shifting its phase,
    and then taking the FFT to shift the signal.

    Parameters
    ----------
    y : 1D `~numpy.ndarray`
        Signal to be shifted. Only 1D supported.
    shift : float
        Amount to shift the signal by in channels.
    pad : bool
        Pad the signal to prevent aliasing.
    window : bool
        Apply a Welch window to prevent ringing.
    nan_treatment : 'fill'
        'fill' replaces NaN values with `fill_value` before the FFT.
    fill_value : float
        Value used to replace NaN values. Used if `nan_treatment=='fill'`.
    keep_nan : bool
        If `True` the output will keep the NaN values. If `False` NaN values will be filled.

    Returns
    -------
    new_y : 1D `~numpy.ndarray`
        Phase shifted version of the original signal `y`.
    """

    nch = len(y)

    # Pad if requested.
    # Padding reduces aliasing, but is more expensivep.
    if pad:
        padded, nskip = fft_pad(y)
    else:
        padded, nskip = y.copy(), 0

    new_len = len(padded)

    phase_shift = 2.0 * np.pi * shift / new_len

    yf = padded.copy()
    nan_mask = np.isnan(padded)

    if nan_treatment == "fill":
        yf[nan_mask] = fill_value
    else:
        return NotImplemented

    # Take the inverse FFT.
    ifft = np.fft.ifft(yf)

    # Split into amplitude and phase.
    amp = abs(ifft)
    phs = np.angle(ifft)

    # Array of indices for computing the phase shift at each channel.
    harr = np.arange(0, new_len // 2)
    # Index the phase shift samples this way to
    # get a Hermite-symmetric complex exponential.
    farr = np.hstack((harr, harr - new_len // 2))

    # Shift and apply window if needed.
    phs += farr * phase_shift
    if window:
        amp *= 1.0 - (farr / float(new_len / 2.0)) ** 2

    # Combine again and take the FFT.
    shifted_ifft = amp * np.cos(phs) + 1j * amp * np.sin(phs)
    shifted_y = np.fft.fft(shifted_ifft)

    # Remove padding and take the real part.
    new_y = shifted_y[nskip : nskip + nch].real
    new_nan_mask = nan_mask[nskip : nskip + nch]

    # Shift NaN elements.
    if keep_nan:
        if abs(shift) < 1:
            new_y[new_nan_mask] = np.nan
        else:
            new_y[np.roll(new_nan_mask, int(shift))] = np.nan

    return new_y


@log_function_call()
def smooth(data, method="hanning", width=1, kernel=None, show=False):
    """
    Smooth or Convolve a spectrum, optionally decimating it.
    A number of methods from astropy.convolution can be selected
    with the method= keyword.

    Default smoothing is hanning.

    Parameters
    ----------
    data : `~numpy.ndarray`
        Input data array to smooth. Note smoothing array does not need a
        WCS since it is channel based.
    method : string, optional
        Smoothing method. Valid are: 'hanning', 'boxcar' and
        'gaussian'. Minimum match applies.
        The default is 'hanning'.
    width : int, optional
        Effective width of the convolving kernel.  Should ideally be an
        odd number.
        For 'hanning' this should be 1, with a 0.25,0.5,0.25 kernel.
        For 'boxcar' an even value triggers an odd one with half the
        signal at the edges, and will thus not reproduce GBTIDL.
        For 'gaussian' this is the FWHM of the final beam. We normally
        assume the input beam has FWHM=1, pending resolution on cases
        where CDELT1 is not the same as FREQRES.
        The default is 1.
    kernel : numpy array, optional
        A numpy array which is the kernel by which the signal is convolved.
        Use with caution, as it is assumed the kernel is normalized to
        one, and is symmetric. Since width is ill-defined here, the user
        should supply an appropriate number manually.
        NOTE: not implemented yet.
        The default is None.
    show : bool, optional
        If set, the kernel is returned, instead of the convolved array.
        The default is False.

    Raises
    ------
    Exception
        If no valid smoothing method is given.

    Returns
    -------
    s : `~numpy.ndarray`
        The new convolved spectrum.

    """
    # note that these methods always return odd number in the kernel
    available_methods = {
        "boxcar": Box1DKernel,  # (e.g. width=2 gives hanning)
        "hanning": Trapezoid1DKernel,  # only for width=1
        "gaussian": Gaussian1DKernel,
    }
    method = minimum_string_match(method, list(available_methods.keys()))
    if method == None:
        raise ValueError(f"Unrecognized input method {method}. Must be one of {list(available_methods.keys())}")
    kernel = available_methods[method](width)
    if show:
        return kernel
    # the boundary='extend' matches  GBTIDL's  /edge_truncate CONVOL() method
    if hasattr(data, "mask"):
        mask = data.mask
    else:
        mask = None
    new_data = convolve(data, kernel, boundary="extend")  # , nan_treatment="fill", fill_value=np.nan, mask=mask)
    return new_data


def data_ishift(y, ishift, axis=-1, remove_wrap=True, fill_value=np.nan):
    """
    Shift `y` by `ishift` channels, where `ishift` is a natural number.

    Parameters
    ----------
    y : array
        Data to be shifted.
    ishift : int
        Amount to shift data by.
    axis : int
        Axis along which to apply the shift.
    remove_wrap : bool
        Replace channels that wrap around with `fill_value`.
    fill_value : float
        Value used to replace the data in channels that wrap around after the shift.

    Returns
    -------
    new_y : array
        Shifted `y`.
    """

    new_y = np.roll(y, ishift, axis=axis)

    if remove_wrap:
        if ishift < 0:
            new_y[ishift:] = fill_value
        else:
            new_y[:ishift] = fill_value

    return new_y


def data_fshift(y, fshift, method="fft", pad=False, window=True):
    """
    Shift `y` by `fshift` channels, where `abs(fshift)<1`.

    Parameters
    ----------
    y : array
        Data to be shifted.
    fshift : float
        Amount to shift the data by.
        abs(fshift) must be less than 1.
    method : "fft" | "interpolate"
        Method to use for shifting.
        "fft" uses a phase shift.
        "interpolate" uses `scipy.ndimage.shift`.
    pad : bool
        Pad the data during the phase shift.
        Only used if `method="fft"`.
    window : bool
        Apply a Welch window during phase shift.
        Only used if `method="fft"`.
    """

    if abs(fshift) > 1:
        raise ValueError("abs(fshift) must be less than one: {fshift}")

    if method == "fft":
        new_y = fft_shift(y, fshift, pad=pad, window=window)
    elif method == "interpolate":
        new_y = ndimage.shift(y, [fshift])

    return new_y


def data_shift(y, s, axis=-1, remove_wrap=True, fill_value=np.nan, method="fft", pad=False, window=True):
    """
    Shift `y` by `s` channels.

    Parameters
    ----------
    y : array
        Data to be shifted.
    s : float
        Amount to shift the data by.
    axis : int
        Axis along which to apply the shift.
    remove_wrap : bool
        Replace channels that wrap around with `fill_value`.
    fill_value : float
        Value used to replace the data in channels that wrap around after the shift.
    method : "fft" | "interpolate"
        Method to use for shifting.
        "fft" uses a phase shift.
        "interpolate" uses `scipy.ndimage.shift`.
    pad : bool
        Pad the data during the phase shift.
        Only used if `method="fft"`.
    window : bool
        Apply a Welch window during phase shift.
        Only used if `method="fft"`.
    """

    ishift = int(np.round(s))  # Integer shift.
    fshift = s - ishift  # Fractional shift.

    logger.debug(f"Shift: s={s}  ishift={ishift} fshift={fshift}")

    if ishift != 0:
        # Apply integer shift.
        y_new = data_ishift(y, ishift, axis=axis, remove_wrap=remove_wrap, fill_value=fill_value)
    else:
        y_new = deepcopy(y)

    if fshift != 0:
        # Apply fractional shift.
        y_new = data_fshift(y_new, fshift, method=method, pad=pad, window=window)

    return y_new



"""
Core utility definitions, classes, and functions
"""

import hashlib
import re
import sys
from pathlib import Path
from typing import Union

# import astropy.units as u
import numpy as np

# import pandas as pd
from astropy.time import Time

ALL_CHANNELS = "all channels"


def select_from(key, value, df):
    """
    Select data where key=value.

    Parameters
    ----------
    key : str
        The key value (SDFITS column name)
    value : any
        The value to match
    df : `~pandas.DataFrame`
        The DataFrame to search

    Returns
    -------
    df : `~pandas.DataFrame`
        The subselected DataFrame

    """
    # nb this fails if value is None
    return df[(df[key] == value)]


def eliminate_flagged_rows(df, flag):
    """
    Remove rows from an index (selection) where all channels have been flagged.

    Parameters
    ----------
    df : `~pandas.DataFrame`
        The input dataframe from which flagged rows will be removed.
    flag : `~pandas.DataFrame`
        The flag dataframe.  Should be the result of e.g. `~util.Flag.final`

    Returns
    -------
        A data frame which is the input data frame with flagged rows removed.
    """
    if len(flag) > 0:
        # in the final flagging selection any rows that have CHAN=ALL_CHANNELS
        # indicate that the entire row is flagged
        ff = flag[flag["CHAN"].isin([ALL_CHANNELS])]
        flagged_rows = set(ff["ROW"])
        if len(flagged_rows) > 0:
            userows = list(set(df["ROW"]) - flagged_rows)
            if len(userows) > 0:
                return df[df["ROW"].isin(userows)]
            else:
                return df.iloc[0:0]  # all rows removed
    return df


def indices_where_value_changes(colname, df):
    """
    Find the `~pandas.DataFrame` indices where the value of the input column name changes.

    Parameters
    ----------
    colname : str
        The column name to query.
    df : `~pandas.DataFrame`
            The DataFrame to search

    Returns
    -------
    indices : ~numpy.ndarray
        The indices of the Dataframe where `colname` changes value.

    """
    # @todo add option to return changing values along with index
    # e.g., [["A",0],["B",125],["C",246]]
    # This is some super panda kung-fu.
    # See https://stackoverflow.com/questions/48673046/get-index-where-value-changes-in-pandas-dataframe-column
    if colname not in df:
        raise KeyError(f"Column {colname} not in input DataFrame")
    # df.shift() shifts the index by one, so we are then comparing df[N] to df[N-1]. This gets us
    # a truth table of where values change.  We filter on colname, then return a list of indices
    # where the value is true. Finally, we squeeze out the empty dimensions of the np array.
    ary = df.ne(df.shift()).filter(items=[colname]).apply(lambda x: x.index[x].tolist()).values
    return np.squeeze(ary, axis=1)


def gbt_timestamp_to_time(timestamp):
    """Convert the GBT sdfits timestamp string format to
    an :class:`~astropy.time.Time` object.  GBT SDFITS timestamps have the form
    YYYY_MM_DD_HH:MM:SS in UTC.

    Parameters
    ----------
    timestamp : str or list-like
        The GBT format timestamp as described above. If str, a Time object containing a single time is returned.
        If list-like, a Time object containing  multiple UTC times is returned.

    Returns
    -------
    time : `~astropy.time.Time`
        The time object
    """
    # convert to ISO FITS format  YYYY-MM-DDTHH:MM:SS(.SSS)
    if isinstance(timestamp, str):
        t = timestamp.replace("_", "-", 2).replace("_", "T")
    else:
        t = [ts.replace("_", "-", 2).replace("_", "T") for ts in timestamp]
    return Time(t, scale="utc")


def generate_tag(values, hashlen, add_time=True):
    """
    Generate a unique tag based on input values.  A hash object is
    created from the input values using SHA256, and a hex representation is created.
    The first `hashlen` characters of the hex string are returned.

    Parameters
    ----------
    values : array-like
        The values to use in creating the hash object
    hashlen : int, optional
        The length of the returned hash string.
    add_time: bool
        Add the time of the call to the values for hash generation.

    Returns
    -------
    tag : str
        The hash string

    """
    if add_time:
        values.append(Time.now().value)
    data = "".join(map(str, values))
    hash_object = hashlib.sha256(data.encode())
    unique_id = hash_object.hexdigest()
    return unique_id[0:hashlen]


def consecutive(data, stepsize=1):
    """Returns the indices of elements in `data`
    separated by less than stepsize separated into
    groups.

    Parameters
    ----------
    data : array
        Array with values to split.
    stepsize : int
        Maximum separation between elements of `data`
        to be considered a single group.

    Returns
    -------
    groups : `~numpy.ndarray`
        Array with values of `data` separated into groups.
    """
    return np.split(data, np.where(np.diff(data) >= stepsize)[0] + 1)


def sq_weighted_avg(a, axis=0, weights=None):
    # @todo make a generic moment or use scipy.stats.moment
    r"""Compute the mean square weighted average of an array (2nd moment).

    :math:`v = \sqrt{\frac{\sum_i{w_i~a_i^{2}}}{\sum_i{w_i}}}`

    Parameters
    ----------
    a : `~numpy.ndarray`
        The data to average
    axis : int
        The axis over which to average the data.  Default: 0
    weights : `~numpy.ndarray` or None
        The weights to use in averaging.  The weights array must be the
        length of the axis over which the average is taken.  Default:
        `None` will use equal weights.

    Returns
    -------
    average : `~numpy.ndarray`
        The average along the input axis
    """
    if weights is None:
        w = np.ones_like(a)
    else:
        w = weights
    v = np.sqrt(np.average(a * a, axis=axis, weights=w))
    return v


def get_project_root() -> Path:
    """
    Returns the project root directory.
    """
    return Path(__file__).parent.parent.parent.parent


def get_project_testdata() -> Path:
    """
    Returns the project testdata directory
    """
    return get_project_root() / "testdata"


def get_project_configuration() -> Path:
    """
    Returns the directory where dysh configuration files are kept.

    Returns
    -------
    Path
        The project configuration directory.

    """
    return get_project_root() / "conf"


def get_size(obj, seen=None):
    """Recursively finds size of objects.
    See https://goshippo.com/blog/measure-real-size-any-python-object/
    """
    size = sys.getsizeof(obj)
    if seen is None:
        seen = set()
    obj_id = id(obj)
    if obj_id in seen:
        return 0
    # Important mark as seen *before* entering recursion to gracefully handle
    # self-referential objects
    seen.add(obj_id)
    if isinstance(obj, dict):
        size += sum([get_size(v, seen) for v in obj.values()])
        size += sum([get_size(k, seen) for k in obj.keys()])
    elif hasattr(obj, "__dict__"):
        size += get_size(obj.__dict__, seen)
    elif hasattr(obj, "__iter__") and not isinstance(obj, (str, bytes, bytearray)):
        size += sum([get_size(i, seen) for i in obj])
    return size


def minimum_string_match(s, valid_strings):
    """
    return the valid string from a list, given a minimum string input

    Example:  minimum_string_match('a',['alpha','beta','gamma'])
    returns:  'alpha'

    Parameters
    ----------
    s : string
        string to use for minimum match
    valid_strings : list of strings
        list of full strings to min match on

    Returns
    -------
    string
        matched string, if one is found.
        Otherwise "None" is returned.

    """
    n = len(valid_strings)
    m = []
    for i in range(n):
        if valid_strings[i].find(s) == 0:
            m.append(i)
    if len(m) == 1:
        return valid_strings[m[0]]
    return None


def uniq(seq):
    """Remove duplicates from a list while preserving order.
    from http://stackoverflow.com/questions/480214/how-do-you-remove-duplicates-from-a-list-in-python-whilst-preserving-order
    """
    seen = set()
    seen_add = seen.add
    return [x for x in seq if x not in seen and not seen_add(x)]


def keycase(d, case="upper"):
    """
    Change the case of dictionary keys

    Parameters
    ----------
    d : dict
        The input dictionary
    case : str, one of 'upper', 'lower'
        Case to change keys to The default is "upper".

    Returns
    -------
    newDict : dict
        A copy of the dictionary with keys changed according to `case`

    """
    if case == "upper":
        newDict = {k.upper(): v for k, v in d.items()}
    elif case == "lower":
        newDict = {k.lower(): v for k, v in d.items()}
    return newDict


# Example of logging a function call
# @log_function_call(log_level="debug")
def powerof2(number):
    """
    Computes the closest power of 2 for a given `number`.

    Parameters
    ----------
    number : float
        number to determine the closest power of 2.

    Returns
    -------
    pow2 : int
        the closest power of 2.
    """

    return round(np.log10(number) / np.log10(2.0))


# From astropy.io.fits.Card:
# FSC commentary card string which must contain printable ASCII characters.
# Note: \Z matches the end of the string without allowing newlines
_ascii_text_re = re.compile(r"[ -~]*\Z")


def _ensure_ascii_str(text: str, check: bool = False) -> str:
    """does the actual cleaning of a text string"""
    clean_text = text.encode("ascii", "ignore").decode("ascii")
    clean_text = clean_text.replace("\n", " ")
    if check and _ascii_text_re.match(clean_text) is None:
        raise ValueError(f"Unable to fully clean string:{clean_text!r} of non-ASCII or non-printable characters.")

    return clean_text


def ensure_ascii(text: Union[str, list[str]], check: bool = False) -> Union[str, list[str]]:
    """
    Remove non-printable ASCII characters from a string or list of strings. This is to ensure that
    FITS cards conform to the standard

    Parameters
    ----------
    text : str
        The text to clean

    check: bool
        Check if the clean value is truly clean according to astropy FITS, raise ValueError if not
    Returns
    -------
    str or list[str]
        The cleaned text

    """
    if isinstance(text, str):
        return _ensure_ascii_str(text)
    else:
        clean_text = []
        for c in text:
            clean_text.append(_ensure_ascii_str(c))
        return clean_text


def convert_array_to_mask(a, length, value=True):
    """
    This method interprets a simple or compound array and returns a numpy mask
    of length `length`. Single arrays/tuples will be treated as element index lists;
    nested arrays will be treated as *inclusive* ranges, for instance:

    ``
    # mask elements 1 and 10
    convert_array_to_mask([1,10])
    # mask elements 1 thru 10 inclusive
    convert_array_to_mask([[1,10]])
    # mask ranges 1 thru 10 and 47 thru 56 inclusive, and element 75
    convert_array_to_mask([[1,10], [47,56], 75)])
    # tuples also work, though can be harder for a human to read
    convert_array_to_mask(((1,10), [47,56], 75))
    ``

    Parameters
    ----------
    a : number or array-like
        The
    length : int
        The length of the mask to return, e.g. the number of channels in a spectrum.

    value : bool
        The value to fill the mask with.  True to mask data, False to unmask.

    Returns
    -------
    mask : ~np.ndarray
        A numpy array where the mask is True according to the rules above.

    """

    if a == ALL_CHANNELS:
        return np.full(length, value)

    mask = np.full(length, False)

    for v in a:
        if isinstance(v, (tuple, list, np.ndarray)) and len(v) == 2:
            # If there are just two numbers, interpret is as an inclusive range
            mask[v[0] : v[1] + 1] = value
        else:
            mask[v] = value
    return mask


def abbreviate_to(length, value, squeeze=True):
    """
    Abbreviate a value for display in limited space. The abbreviated
    value will have initial characters, ellipsis, and final characters, e.g.
    '[(a,b),(c,d)...(w,x),(y,z)]'.

    Parameters
    ----------
    length : int
        Maximum string length.
    value : any
        The value to be abbreviated.
    squeeze : bool, optional
        Squeeze blanks. If True, replace ", " (comma space) with "," (comma). The default is True.

    Returns
    -------
    strv : str
        Abbreviated string representation of the input value

    """
    strv = str(value)
    if squeeze:
        strv = strv.replace(", ", ",")
    if len(strv) > length:
        bc = int(length / 2) - 1
        ec = bc - 1
        strv = strv[0:bc] + "..." + strv[-ec:]
    return strv



import numbers
import warnings
from collections.abc import Sequence
from copy import deepcopy

import astropy.units as u
import numpy as np
import pandas as pd
from astropy.coordinates import Angle
from astropy.table import Table
from astropy.time import Time
from astropy.units.quantity import Quantity
from pandas import DataFrame

from ..log import logger

# from ..fits import default_sdfits_columns
from . import ALL_CHANNELS, abbreviate_to, gbt_timestamp_to_time, generate_tag, keycase

default_aliases = {
    "freq": "crval1",
    "ra": "crval2",
    "dec": "crval3",
    "glon": "crval2",
    "glat": "crval3",
    "gallon": "crval2",
    "gallat": "crval3",
    "elevation": "elevatio",
    "source": "object",
    "pol": "plnum",
    "subref": "subref_state",  # subreflector state
}

# pd.options.mode.copy_on_write = True


# workaround to avoid circular import error in sphinx (and only sphinx)
def _default_sdfits_columns():
    from ..fits import default_sdfits_columns

    return default_sdfits_columns()


DEFAULT_COLUMN_WIDTH = 32  # char
DEFAULT_COLUMN_TYPE = f"<U{DEFAULT_COLUMN_WIDTH}"


class SelectionBase(DataFrame):
    """This class is the base class for selection and flagging. Selection and flagging are both kinds of
    data selection, so `SelectionBase` can encapsulate most necessary functionality.  Derived classes
    implement specific named methods e.g. select_channel, flag_channel that will simply call
    the base class methods.
    """

    def __init__(self, initobj, aliases=default_aliases, **kwargs):
        if hasattr(initobj, "_index"):  # it's an SDFITSLoad object
            super().__init__(initobj._index)
            DEFKEYS = list(initobj._index.keys())
        else:
            super().__init__(initobj)  # it's a Selection or DataFrame
            DEFKEYS = _default_sdfits_columns()
        # adding attributes that are not columns will result
        # in a UserWarning, which we can safely ignore.
        warnings.simplefilter("ignore", category=UserWarning)
        self._add_utc_column()
        self["CHAN"] = None
        # if we want Selection to replace _index in sdfits
        # construction this will have to change. if hasattr("_index") etc
        self._idtag = ["ID", "TAG"]
        DEFKEYS.extend(["FITSINDEX", "CHAN", "UTC", "# SELECTED"])
        # add ID and TAG as the first columns
        for i in range(len(self._idtag)):
            DEFKEYS.insert(i, self._idtag[i])
        # add channel, astropy-based timestamp, and number rows selected
        DEFKEYS = np.array(DEFKEYS)
        # set up object types for the np.array
        dt = np.full(len(DEFKEYS) - 1, np.dtype(DEFAULT_COLUMN_TYPE))
        dt[0] = np.int32
        # add number selected column which is an int
        dt = np.insert(dt, len(dt), np.int32)
        # ID is also an int
        dt[0] = np.int32
        self._defkeys = DEFKEYS
        self._deftypes = dt
        self._make_table()
        self._valid_coordinates = ["RA", "DEC", "GALLON", "GALLAT", "GLON", "GLAT", "CRVAL2", "CRVAL3"]
        self._selection_rules = {}
        self._aliases = {}
        self.alias(aliases)
        self._channel_selection = None
        self._flag_channel_selection = {}  # used in Flag only
        warnings.resetwarnings()

    def _add_utc_column(self):
        """
        Add column to the selection/flag dataframe with a
        representation of the SDFITS UTC timestamp, which is a string,
        as an ~astropy.time.Time.

        Returns
        -------
        None.
        """
        self["UTC"] = gbt_timestamp_to_time(self.TIMESTAMP)

    def _make_table(self):
        """Create the table for displaying the selection rules"""
        self._table = Table(data=None, names=self._defkeys, dtype=self._deftypes)
        for t in self._idtag:
            self._table.add_index(t)

    @property
    def aliases(self):
        """
        The aliases that may be used to refer to SDFITS columns.

        Returns
        -------
        dict
            The dictionary of aliases and SDFITS column names
        """
        return self._aliases

    def alias(self, aliases):
        """
        Alias a set of keywords to existing columns. Multiple aliases for
        a single column are allowed, e.g.,
        { 'glon':'crval2', 'lon':'crval2'}

        Parameters
        ----------
        aliases : {}
            The dictionary of keywords and column names
            where the new alias is the key and
            the column name is the value and , i.e., {alias:column}

        Returns
        -------
        None.

        Raises
        ------
            ValueError if the column name is not recognized.
        """
        self._check_keys(aliases.values())
        for k, v in aliases.items():
            self._alias(k, v)

    def _alias(self, key, column):
        """
        Alias a new keyword to an existing column, e.g..
        to alias the SDFITS column 'CRVAL2' as 'RA':

            `alias('RA','CRVAL2')`

        The map is case insensitive, so `alias('ra', 'crval2')` also works.

        Parameters
        ----------
        key : str
            The new keyword to use as an alias.
        column : str
            The existing SDFITS column name to alias

        Returns
        -------
        None.

        """
        self._aliases[key.upper()] = column.upper()

    def _set_pprint_exclude_names(self):
        """Use `~astropy.Table.pprint_exclude_names` to set the list
        columns that have no entries.
        """
        if len(self._table) > 0:
            emptycols = np.array(self._table.colnames)[
                [np.all([self._table[k].data[i] == "" for i in range(len(self._table))]) for k in self._table.colnames]
            ]
            self._table.pprint_exclude_names.set(emptycols)

    def columns_selected(self):
        """The names of any columns which were used in a selection rule

        Returns
        -------
        colnames - set
            A set of str column names. An empty set is returned if no selection rule has yet been made.
        """
        if len(self._table) == 0:
            return set()

        self._set_pprint_exclude_names()  # ensure __attributes__ gets set.
        return (
            set(self._table.colnames)
            - set(self._table.meta["__attributes__"]["pprint_exclude_names"])
            - set(["# SELECTED", "ID", "TAG"])
        )

    def _sanitize_input(self, key, value):
        """
        Sanitize a key-value pair for. Coordinate types are checked for.

        Parameters
        ----------
        key : str
            upper case key value

        value : any
            The value for the key

        Returns
        -------
        sanitized_value : str
            The sanitized value
        """
        # @todo   Allow minimum match str for key?
        if key in self._aliases.keys():
            key = self._aliases[key]
        if key not in self:
            raise KeyError(f"{key} is not a recognized column name.")
        v = self._sanitize_coordinates(key, value)
        # deal with Time here or later?
        self._check_for_disallowed_chars(key, value)
        return v

    def _sanitize_coordinates(self, key, value):
        """
        Sanitize a coordinate selection key-value pair. Coordinates will be
        converted to floats before the final value is created.

        Parameters
        ----------
        key : str
            upper case key value

        value : any
            The value for the key.  It can be a single float,
            a single Angle (Quantity), a tuple of Angles
            (a1,a2,a3) or an Angle tuple, e.g., (n1,n2)*u.degree

        Returns
        -------
        sanitized_value : str
            The sanitized value.
        """
        if key not in self._valid_coordinates and key not in self.aliases:
            return value
        # note Quantity is derivative of np.ndarray, so
        # need to filter that out in the recursive call.
        # This is to handle (q1,q2) as a range.
        # (n1,n2)*u.degree is handled below
        if isinstance(value, (tuple, np.ndarray, list)) and not isinstance(value, Quantity):
            return [self._sanitize_coordinates(key, v) for v in value]
        if isinstance(value, numbers.Number):
            a = Angle(value * u.degree)
        else:  # it should be a str or Quantity
            a = Angle(value)
        return a.degree

    def _check_for_disallowed_chars(self, key, value):
        # are there any?  coordinates will already
        # be transformed to decimal degrees
        pass

    def _generate_tag(self, values, hashlen=9):
        """
        Generate a unique tag based on row values.  A hash object is
        created from the input values using SHA256, and a hex representation is created.
        The first `hashlen` characters of the hex string are returned.

        Parameters
        ----------
        values : array-like
            The values to use in creating the hash object
        hashlen : int, optional
            The length of the returned hash string. The default is 9.

        Returns
        -------
        tag : str
            The hash string

        """
        return generate_tag(values, hashlen)

    @property
    def _next_id(self) -> int:
        """
        Get the next ID number in the table.

        Returns
        -------
        id : int
            The highest existing ID number plus one
        """
        ls = len(self._table)
        if ls == 0:
            return 0
        return max(self._table["ID"]) + 1

    def _check_keys(self, keys):
        """
        Check a dictionary for unrecognized keywords.  This method is called in any select method to check inputs.

        Parameters
        ----------
        keys : list or array-like
           Keyword arguments

        Returns
        -------
        None.

        Raises
        ------
        KeyError
            If one or more keywords are unrecognized

        """
        # ignorekeys = ["PROPOSED_CHANNEL_RULE"]
        unrecognized = []
        ku = [k.upper() for k in keys]
        # for k in ignorekeys:
        #    if k in ku:
        #        ku.remove(k)
        for k in ku:
            if k not in self and k not in self._aliases:
                unrecognized.append(k)
        if len(unrecognized) > 0:
            raise KeyError(f"The following keywords were not recognized: {unrecognized}")

    def _check_numbers(self, **kwargs):
        self._check_type(numbers.Number, "Expected numeric value for these keywords but did not get a number", **kwargs)

    def _check_range(self, **kwargs):
        bad = []
        badtime = []
        for k, v in kwargs.items():
            ku = k.upper()
            if not isinstance(v, (tuple, list, np.ndarray)):
                raise ValueError(f"Invalid input for key {ku}={v}. Range inputs must be tuple or list.")
            for a in v:
                if a is not None:
                    if isinstance(a, Quantity):
                        a = self._sanitize_coordinates(ku, a)
                    try:
                        if ku == "UTC":
                            badtime = self._check_type(Time, "Expected Time", silent=True, **{ku: a})
                        else:
                            self._check_numbers(**{ku: a})
                    except ValueError:
                        bad.append(ku)
        if len(bad) > 0 or len(badtime) > 0:
            msg = "Expected"
            a = " "
            if len(badtime) > 0:
                msg += f" Time object for {badtime}"
                a = " and "
            if len(bad) > 0:
                msg += a
                msg += f"numeric value(s) for {bad} "
            msg += " but did not get that."
            raise ValueError(msg)

    def _check_type(self, reqtype, msg, silent=False, **kwargs):
        # @todo allow Quantities
        """
        Check that a list of keyword arguments is all a specified type.

        Parameters
        ----------

        reqtype : type
            The object type to check against, e.g. numbers.Number, str, etc

        msg : str
            The exception message to show if the inputs are not the specific reqtype

        **kwargs : dict or key=value
           Keyword arguments

        Raises
        ------
        ValueError
            If one or more of the values is not numeric.

        Returns
        -------
        None or, if silent is True,s a list of keywords that raised errors.

        """
        # deal with potential arrays first by calling
        # this method recursively on each array member.
        kw = deepcopy(kwargs)  # prevent concurrent modification
        recursive_bad = []
        for k, v in kwargs.items():
            # if the input value is an array, then we want to
            # check the type for each member of the array, but
            # not raise an exception (silent=True) but collect
            # the bad ones and pass them on.
            if isinstance(v, (Sequence, np.ndarray)) and not isinstance(v, str):
                bad = [self._check_type(reqtype, msg, **{k: x}, silent=True) for x in v]
                # there's probably a smarter way to do this
                for i in range(len(bad)):
                    if len(bad[i]) != 0:
                        recursive_bad.extend(bad[i])
                kw.pop(k)
        # now the main check
        ku = np.ma.masked_array([k.upper() for k in kw.keys()])
        ku.mask = np.array([isinstance(x, reqtype) for x in kw.values()])
        if len(recursive_bad) != 0:
            ku = np.ma.append(ku, recursive_bad)
        if silent:
            return list(ku[~ku.mask])
        if not np.all(ku.mask):
            raise ValueError(f"{msg}: {np.squeeze(ku[~ku.mask])}")

    def _check_for_duplicates(self, df):
        """
        Check that the user hasn't already added a rule matching this one

        Parameters
        ----------
        df : ~pandas.DataFrame
            The selection to check


        Returns
        -------
        bool
           True if a duplicate was found, False if not.

        """
        for _id, s in self._selection_rules.items():
            tag = self._table.loc[_id]["TAG"]
            if s.equals(df):
                tag = self._table.loc[_id]["TAG"]
                warnings.warn(
                    f"A rule that results in an identical selection has already been added: ID: {_id}, TAG:{tag}."
                    " Ignoring."
                )
                return True
        return False

    def _addrow(self, row, dataframe, tag=None):
        """
        Common code to add a tagged row to the internal table after the selection has been created.
        Should be called in select* methods.

        Parameters
        ----------
        row : dict
            key, value pairs of the selection
        dataframe : ~pandas.DataFrame
            The dataframe created by the selection.
        tag : str, optional
            An identifying tag by which the rule may be referred to later.
            If None, a  randomly generated tag will be created.
        Returns
        -------
        None.

        """
        if self._check_for_duplicates(dataframe):
            return
        if tag is not None:
            row["TAG"] = tag
        else:
            gentag = []
            # guarantee a unique seed by
            # including relevant key=value, which will be unique
            for k, v in row.items():
                if v is not None and v != "":
                    gentag.append(f"{k}={v}")
            row["TAG"] = self._generate_tag(gentag)
        row["ID"] = self._next_id
        row["# SELECTED"] = len(dataframe)
        self._selection_rules[row["ID"]] = dataframe
        for k, v in row.items():
            row[k] = abbreviate_to(DEFAULT_COLUMN_WIDTH, v)
        self._table.add_row(row)
        # self._last_row_added = row

    def _base_select(self, tag=None, **kwargs):
        """Add one or more exact selection/flag rules, e.g., `key1 = value1, key2 = value2, ...`
        If `value` is array-like then a match to any of the array members will be selected/flagged.
        Derived classes will call this method with their own specific name, i.e. `select` or `flag`.

        Parameters
        ----------
            tag : str
                An identifying tag by which the rule may be referred to later.
                If None, a  randomly generated tag will be created.
            key : str
                The key  (SDFITS column name or other supported key)
            value : any
                The value to select

        Returns
        -------
            True if the selection resulted in a new fule, False if not (no data selected)

        """
        # pop these before check_keys, which is intended to check SDFITS keywords
        proposed_channel_rule = kwargs.pop("proposed_channel_rule", None)
        # if called via _select_from_mixed_kwargs, then we want to merge all the
        # selections
        df = kwargs.pop("startframe", self)
        self._check_keys(kwargs.keys())
        row = {}

        single_value_queries = None
        multi_value_queries = None
        for k, v in list(kwargs.items()):
            ku = k.upper()
            if ku in self._aliases:
                ku = self._aliases[ku]
            v = self._sanitize_input(ku, v)
            # If a list is passed in, it must be composed of strings.
            # Numeric lists are intepreted as ranges, so must be
            # selected by user with select_range
            if isinstance(v, (Sequence, np.ndarray)) and not isinstance(v, str):
                query = None
                for vv in v:
                    if ku == "UTC":
                        self._check_type(
                            Time,
                            "Expected Time object but got something else.",
                            **{ku: vv},
                        )
                    # if it is a string, then OR them.
                    # e.g. object = ["NGC123", "NGC234"]
                    if isinstance(vv, str):
                        thisq = f'{ku} == "{vv}"'
                    else:
                        thisq = f"{ku} == {vv}"
                    if query is None:
                        query = thisq
                    else:
                        query += f"| {thisq}"
                    # for pd.merge to give the correct answer, we would
                    # need "inner" on the first one and "outer" on subsequent
                    # df = pd.merge(df, df[df[ku] == vv], how="inner")
                if multi_value_queries is None:
                    multi_value_queries = f"({query})"
                else:
                    multi_value_queries += f"&({query})"
            else:
                if isinstance(v, str):
                    thisq = f'{ku} == "{v}"'
                else:
                    thisq = f"{ku} == {v}"
                if single_value_queries is None:
                    single_value_queries = thisq
                else:
                    single_value_queries += f"& {thisq}"
            row[ku] = v
        if multi_value_queries is not None and single_value_queries is not None:
            query = f"{multi_value_queries} & {single_value_queries}"
        elif multi_value_queries is None and single_value_queries is not None:
            query = single_value_queries
        elif multi_value_queries is not None and single_value_queries is None:
            query = multi_value_queries
        else:
            warnings.warn("There was no data selection")  # should never happen
            return False
        df = df.query(query)
        if df.empty:
            warnings.warn("Your selection rule resulted in no data being selected. Ignoring.")
            return False
        df.loc[:, "CHAN"] = proposed_channel_rule  # this column is normally None so no need to check if None first.
        self._addrow(row, df, tag)
        return True

    def _base_select_range(self, tag=None, **kwargs):
        """
        Select a range of inclusive values for a given key(s).
        e.g., `key1 = (v1,v2), key2 = (v3,v4), ...`
        will select data  `v1 <= data1 <= v2, v3 <= data2 <= v4, ... `
        Upper and lower limits may be given by setting one of the tuple values
        to None. e.g., `key1 = (None,v1)` for an upper limit `data1 <= v1` and
        `key1 = (v1,None)` for a lower limit `data >=v1`.  Lower
        limits may also be specified by a one-element tuple `key1 = (v1,)`.

        Parameters
        ----------
        tag : str, optional
            An identifying tag by which the rule may be referred to later.
            If None, a  randomly generated tag will be created.
        key : str
            The key (SDFITS column name or other supported key)
        value : array-like
            Tuple or list giving the lower and upper limits of the range.

        Returns
        -------
        None.

        """
        self._check_keys(kwargs.keys())
        self._check_range(**kwargs)
        row = {}
        df = self
        for k, v in list(kwargs.items()):
            ku = k.upper()
            if ku in self._aliases:
                ku = self._aliases[ku]
            v = self._sanitize_input(ku, v)
            # deal with a tuple quantity
            if isinstance(v, Quantity):
                v = v.value
            vn = []
            # deal with quantity inside a tuple.
            for q in v:
                # ultimately will need a map of
                # desired units, so e.g. if
                # GHz used, then the value is expressed in Hz
                if isinstance(q, Quantity):
                    vn.append(q.value)
                else:
                    vn.append(q)
            v = vn
            row[ku] = v
            if len(v) == 2:
                if v[0] is not None and v[1] is not None:
                    df = pd.merge(df, df[(df[ku] <= v[1]) & (df[ku] >= v[0])], how="inner")
                elif v[0] is None:  # upper limit given
                    df = pd.merge(df, df[(df[ku] <= v[1])], how="inner")
                else:  # lower limit given (v[1] is None)
                    df = pd.merge(df, df[(df[ku] >= v[0])], how="inner")
            elif len(v) == 1:  # lower limit given
                df = pd.merge(df, df[(df[ku] >= v[0])], how="inner")
            else:
                raise Exception(f"Couldn't parse value tuple {v} for key {k} as a range.")
        if df.empty:
            warnings.warn("Your selection rule resulted in no data being selected. Ignoring.")
            return
        self._addrow(row, df, tag)

    def _base_select_within(self, tag=None, **kwargs):
        """
        Select a value within a plus or minus for a given key(s).
        e.g. `key1 = [value1,epsilon1], key2 = [value2,epsilon2], ...`
        Will select data
        `value1-epsilon1 <= data1 <= value1+epsilon1,`
        `value2-epsilon2 <= data2 <= value2+epsilon2,...`

        Parameters
        ----------
        tag : str, optional
            An identifying tag by which the rule may be referred to later.
            If None, a  randomly generated tag will be created.
        key : str
            The key (SDFITS column name or other supported key)
        value : array-like
            Tuple or list giving the value and epsilon

        Returns
        -------
        None.

        """
        # This is just a type of range selection.
        kw = {}
        for k, v in kwargs.items():
            v1 = v[0] - v[1]
            v2 = v[0] + v[1]
            kw[k] = (v1, v2)
        self._base_select_range(tag, **kw)

    def _base_select_channel(self, channel, tag=None):
        """
        Select channels and/or channel ranges. These are NOT used in :meth:`final`
        but rather will be used to create a mask for calibration or
        flagging. Single arrays/tuples will be treated as channel lists;
        nested arrays will be treated as  *inclusive* ranges. For instance:

        ``
        # select channel 24
        select_channel(24)
        # selects channels 1 and 10
        select_channel([1,10])
        # selects channels 1 thru 10 inclusive
        select_channel([[1,10]])
        # select channel ranges 1 thru 10 and 47 thru 56 inclusive, and channel 75
        select_channel([[1,10], [47,56], 75)])
        # tuples also work, though can be harder for a human to read
        select_channel(((1,10), [47,56], 75))
        ``

        *Note* : channel numbers start at zero.

        Parameters
        ----------
        channel : number, or array-like
            The channels to select

        Returns
        -------
        None.

        """
        # We don't want to get into trying to merge
        # different, possibly exclusive, channel selections.
        # This also avoids the side effect of using self to
        # compute "# Selected" in _addrow
        if self._channel_selection is not None:
            raise Exception(
                "You can only have one channel selection rule. Remove the old rule before creating a new one."
            )
        self._check_numbers(chan=channel)
        if isinstance(channel, numbers.Number):
            channel = [int(channel)]
        self._channel_selection = channel
        self._addrow({"CHAN": abbreviate_to(DEFAULT_COLUMN_WIDTH, channel)}, dataframe=self, tag=tag)

    # NB: using ** in doc here because `id` will make a reference to the
    # python built-in function.  Arguably we should pick a different
    # keyword but 'id' is easy for user.
    def remove(self, id=None, tag=None):
        """Remove (delete) a selection rule(s).
        You must specify either **id** or **tag** but not both. If there are
        multiple rules with the same tag, they will all be deleted.

        Parameters
        ----------
            id : int
                The ID number of the rule as displayed in `show()`
            tag : str
                An identifying tag by which the rule may be referred to later.
        """
        if id is not None and tag is not None:
            raise Exception("You can only specify one of id or tag")
        if id is None and tag is None:
            raise Exception("You must specify either id or tag")
        if id is not None:
            if id in self._selection_rules:
                # We will assume that selection_rules and table
                # have been kept in sync.  The implementation
                # should ensure this.
                del self._selection_rules[id]
                row = self._table.loc_indices["ID", id]
                # there is only one row per ID
                self._table.remove_row(row)
            else:
                raise KeyError(f"No ID = {id} found in this Selection")
        else:
            # need to find IDs of selection rules where TAG == tag.

            # This will raise keyerror if tag not matched, so no need
            # to raise our own, unless we want to change the messgae.
            matching_indices = self._table.loc_indices["TAG", tag]
            #   raise KeyError(f"No TAG = {tag} found in this Selection")
            matching = Table(self._table[matching_indices])
            for i in matching["ID"]:
                del self._selection_rules[i]
                # self._selection_rules.pop(i, None) # also works
            self._table.remove_rows(matching_indices)

    def clear(self):
        """Remove all selection rules"""
        self._selection_rules = {}
        self._flag_channel_selection = {}
        self._make_table()
        self._channel_selection = None

    def show(self):
        """
        Print the current selection rules. Only columns with a rule are shown.
        The first two columns are ID number a TAG string. Either of these may be used
        to :meth:`remove` a row.  The final column `# SELECTED` gives
        the number of rows that a given rule selects from the original.
        The :meth:`final` selection may be fewer rows because each selection rule
        is logically OR'ed to create the final selection.

        Returns
        -------
        None.

        """
        self._set_pprint_exclude_names()
        print(self._table)

    @property
    def final(self):
        """
        Create the final selection. This is done by a logical AND of each
        of the selection rules (specifically `pandas.merge(how='inner')`).

        Returns
        -------
        final : DataFrame
            The resultant selection from all the rules.
        """
        return self.merge(how="inner")

    def merge(self, how, on=None):
        """
        Merge selection rules using a specific
        type of join.

        Parameters
        ----------
        how : {‘left’, ‘right’, ‘outer’, ‘inner’, ‘cross’}, no default.
            The type of join to be performed. See :meth:`pandas.merge()`.

        on: label or list
            Column or index level names to join on. These must be found in both DataFrames.
            If on is None and not merging on indexes then this defaults to the intersection
            of the columns in both DataFrames.

        Returns
        -------
        final : DataFrame
            The resultant selection from all the rules.

        """
        if len(self._selection_rules.values()) == 0:
            # warnings.warn("Selection.merge(): upselecting now")
            return DataFrame()
        final = None
        for df in self._selection_rules.values():
            if final is None:
                # need a deepcopy here in case there
                # is only one selection rule, because
                # we don't want to return a reference to the rule
                # which the receiver might modify.
                final = deepcopy(df)
            else:
                final = pd.merge(final, df, how=how, on=on)
        return final

    def _select_from_mixed_kwargs(self, **kwargs):
        """
        Called by calibration routines which may be mixing channel selections
        and exact selections, but **not** 'within' or 'range' selections.

        Parameters
        ----------
        **kwargs : dict
            Keyword arguments.  key=value as in the public selection methods.

        Returns
        -------
        None.

        """

        # get the tag if given or generate one if not
        kwlist = list(kwargs.items())
        tag = kwargs.pop("tag", self._generate_tag(kwlist))
        if len(kwargs) == 0:
            return  # user gave no additional kwargs
        if tag is None:  # in case user did tag=None (facepalm)
            tag = self._generate_tag(kwlist)
        logger.debug(f"working TAG IS {tag}")
        # in order to pop channel we need to check case insensitively
        ukwargs = keycase(kwargs)
        chan = ukwargs.pop("CHANNEL", None)
        if chan is not None:
            self._base_select_channel(chan, tag)
        if len(ukwargs) != 0:
            logger.debug(f"selection {ukwargs}")
            self._base_select(**ukwargs, tag=tag)

    def __deepcopy__(self, memo):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=UserWarning)
            cls = self.__class__
            result = cls.__new__(cls)
            memo[id(self)] = result
            for k, v in self.__dict__.items():
                setattr(result, k, deepcopy(v, memo))
            result._table = self._table.copy()
        return result

    def get(self, key):
        """Get the selection/flag rule by its ID

        Parameters
        ----------
        key : int
            The ID value.  See :meth:`show`.

        Returns
        -------
        ~pandas.DataFrame
            The selection/flag rule
        """
        return self._selection_rules[key]


class Selection(SelectionBase):
    """This class contains the methods for creating rules to select data from an SDFITS object.
    Data (rows) can be selected using any column name in the input SDFITS object.
    Exact selection, range selection, upper/lower limit selection, and any-of selection
    are all supported.

    Users create *selection rules* by specifying keyword (SDFITS columns) and value(s) to be selected.
    Briefly, the selection methods are:

         :meth:`select` - Select exact values

         :meth:`select_range` - Select ranges of values

         :meth:`select_within` - Select a value +/- epsilon

         :meth:`select_channel` - Select channels or ranges of channels

    The Selection object maintains a DataFrame for each selection rule created by the user. The
    :meth:`final` selection is the logical OR of these rules. Users can examine the current selections
    with :meth:`show` which will show the current
    rules and how many rows each rule selects from the unfiltered data.


    Aliases of keywords are supported. The user may add an alias for an existing SDFITS column with :meth:`alias`.   Some default :meth:`aliases` have been defined.
    """

    def select(self, tag=None, **kwargs):
        """Add one or more exact selection rules, e.g., `key1 = value1, key2 = value2, ...`
        If `value` is array-like then a match to any of the array members will be selected.
        For instance `select(object=['3C273', 'NGC1234'])` will select data for either of those
        objects and `select(ifnum=[0,2])` will select IF number 0 or IF number 2.

        Parameters
        ----------
            tag : str
                An identifying tag by which the rule may be referred to later.
                If None, a  randomly generated tag will be created.
            key : str
                The key  (SDFITS column name or other supported key)
            value : any
                The value to select

        """
        self._base_select(tag, **kwargs)

    def select_range(self, tag=None, **kwargs):
        """
        Select a range of inclusive values for a given key(s).
        e.g., `key1 = (v1,v2), key2 = (v3,v4), ...`
        will select data  `v1 <= data1 <= v2, v3 <= data2 <= v4, ... `
        Upper and lower limits may be given by setting one of the tuple values
        to None. e.g., `key1 = (None,v1)` for an upper limit `data1 <= v1` and
        `key1 = (v1,None)` for a lower limit `data >=v1`.  Lower
        limits may also be specified by a one-element tuple `key1 = (v1,)`.

        Parameters
        ----------
        tag : str, optional
            An identifying tag by which the rule may be referred to later.
            If None, a  randomly generated tag will be created.
        key : str
            The key (SDFITS column name or other supported key)
        value : array-like
            Tuple or list giving the lower and upper limits of the range.

        Returns
        -------
        None.

        """
        self._base_select_range(tag, **kwargs)

    def select_within(self, tag=None, **kwargs):
        """
        Select a value within a plus or minus for a given key(s).
        e.g. `key1 = [value1,epsilon1], key2 = [value2,epsilon2], ...`
        Will select data
        `value1-epsilon1 <= data1 <= value1+epsilon1,`
        `value2-epsilon2 <= data2 <= value2+epsilon2,...`

        Parameters
        ----------
        tag : str, optional
            An identifying tag by which the rule may be referred to later.
            If None, a  randomly generated tag will be created.
        key : str
            The key (SDFITS column name or other supported key)
        value : array-like
            Tuple or list giving the value and epsilon

        Returns
        -------
        None.

        """
        self._base_select_within(tag, **kwargs)

    def select_channel(self, chan, tag=None):
        """
        Select channels and/or channel ranges. These are NOT used in :meth:`final`
        but rather will be used to create a mask for calibration or
        flagging. Single arrays/tuples will be treated as channel lists;
        nested arrays will be treated as  *inclusive* ranges. For instance:

        ``
        # select channel 24
        select_channel(24)
        # selects channels 1 and 10
        select_channel([1,10])
        # selects channels 1 thru 10 inclusive
        select_channel([[1,10]])
        # select channel ranges 1 thru 10 and 47 thru 56 inclusive, and channel 75
        select_channel([[1,10], [47,56], 75)])
        # tuples also work, though can be harder for a human to read
        select_channel(((1,10), [47,56], 75))
        ``

        *Note* : channel numbers start at zero.

        Parameters
        ----------
        chan : number, or array-like
            The channels to select

        Returns
        -------
        None.

        """
        self._base_select_channel(chan, tag)


class Flag(SelectionBase):
    """This class contains the methods for creating rules to flag data from an SDFITS object.
    Data (rows) can be selected for flagging using any column name in the input SDFITS object.
    Exact selection, range selection, upper/lower limit selection, and any-of selection
    are all supported.

    Users create *flag rules* by specifying keyword (SDFITS columns) and value(s) to be flagged.
    Briefly, the flag methods are:

         :meth:`flag` - Flag exact values

         :meth:`flag_range` - Flag ranges of values

         :meth:`flag_within` - Flag a value +/- epsilon

         :meth:`flag_channel` - Flag channels or ranges of channels

    The Flag object maintains a DataFrame for each flag rule created by the user. The
    :meth:`final` flag is the logical OR of these rules. Users can examine the current flags
    with :meth:`show` which will show the current
    rules and how many rows each rule selects for flagging from the unfiltered data.

    The actual flags, which are per channel, are stored in the GBTFITSLoad object,
    not in the Flag object. The Flag object just contains the flagging rules.

    Aliases of keywords are supported. The user may add an alias for an existing SDFITS column with :meth:`alias`.   Some default :meth:`aliases` have been defined.

    GBTIDL Flags can be read in with :meth:`read`.
    """

    def flag(self, tag=None, **kwargs):
        """Add one or more exact flag rules, e.g., `key1 = value1, key2 = value2, ...`
        If `value` is array-like then a match to any of the array members will be flagged.
        For instance `flag(object=['3C273', 'NGC1234'])` will select data for either of those
        objects and `flag(ifnum=[0,2])` will flag IF number 0 or IF number 2.  Channels for selected data
        can be flagged using keyword `channel`, e.g., `flag(object='MBM12',channel=[0,23])`
        will flag channels 0 through 23 *inclusive* for object MBM12.

        Parameters
        ----------
            tag : str
                An identifying tag by which the rule may be referred to later.
                If None, a  randomly generated tag will be created.
            key : str
                The key  (SDFITS column name or other supported key)
            value : any
                The value to select

        """
        chan = kwargs.pop("channel", None)
        if chan is not None:
            if isinstance(chan, numbers.Number):
                chan = [int(chan)]
            self._check_numbers(chan=chan)
        if len(kwargs) == 0:
            # The user only entered channel as a keyword, so just call flag_channel
            self.flag_channel(channel=chan, tag=tag)
        else:
            # Select on the other kwargs then add channel to it.
            # Since we are allowing the behavior that the user can select
            # identical rows with different channel flags, we must
            # use a 'proposed channel rule' because the "CHAN" column is not normally set
            # before the _check_for_duplicates call inside _base_select.
            # The selection rules dataframes are allowed to be identical if the
            # the CHAN columns will be different.
            if chan is None:
                kwargs["proposed_channel_rule"] = ALL_CHANNELS
            else:
                kwargs["proposed_channel_rule"] = str(chan)
            success = self._base_select(tag, **kwargs)  # don't do this unless chan input is good.
            if not success:
                return
            idx = len(self._table) - 1
            if chan is not None:
                self._table[idx]["CHAN"] = abbreviate_to(DEFAULT_COLUMN_WIDTH, chan)
                self._flag_channel_selection[idx] = chan
                self._selection_rules[idx]["CHAN"] = str(chan)
            else:
                self._flag_channel_selection[idx] = ALL_CHANNELS
                self._selection_rules[idx].loc[:, "CHAN"] = ALL_CHANNELS

    def flag_channel(self, channel, tag=None, **kwargs):
        """
        Flag  channels and/or channel ranges for *all data*. These are NOT used in :meth:`final`
        but rather will be used to create a mask for
        flagging. Single arrays/tuples will be treated as *channel lists;
        nested arrays will be treated as  *inclusive* ranges. For instance:

        ``
        # flag channel 24
        flag_channel(24)
        # flag channels 1 and 10
        flag_channel([1,10])
        # flags channels 1 thru 10 inclusive
        flag_channel([[1,10]])
        # flag channel ranges 1 thru 10 and 47 thru 56 inclusive, and channel 75
        flag_channel([[1,10], [47,56], 75)])
        # tuples also work, though can be harder for a human to read
        flag_channel(((1,10), [47,56], 75))
        ``

        *Note* : channel numbers start at zero


         Parameters
        ----------
        channel : number, or array-like
            The channels to flag

        Returns
        -------
        None.

        """
        # okay to use base method because we are flagging all rows
        self._base_select_channel(channel, tag, **kwargs)
        idx = len(self._table) - 1
        self._flag_channel_selection[idx] = channel
        self._selection_rules[idx]["CHAN"] = str(channel)
        self._channel_selection = None  # unused for flagging

    def flag_range(self, tag=None, **kwargs):
        """Flag a range of inclusive values for a given key(s).
        e.g., `key1 = (v1,v2), key2 = (v3,v4), ...`
        will flag data  `v1 <= data1 <= v2, v3 <= data2 <= v4, ... `
        Upper and lower limits may be given by setting one of the tuple values
        to None. e.g., `key1 = (None,v1)` for an upper limit `data1 <= v1` and
        `key1 = (v1,None)` for a lower limit `data >=v1`.  Lower
        limits may also be specified by a one-element tuple `key1 = (v1,)`.

        Parameters
        ----------
        tag : str, optional
            An identifying tag by which the rule may be referred to later.
            If None, a  randomly generated tag will be created.
        key : str
            The key (SDFITS column name or other supported key)
        value : array-like
            Tuple or list giving the lower and upper limits of the range.

        Returns
        -------
        None.
        """
        self._base_select_range(tag, **kwargs)
        idx = len(self._table) - 1
        self._flag_channel_selection[idx] = ALL_CHANNELS
        self._selection_rules[idx]["CHAN"] = ALL_CHANNELS
        self._channel_selection = None  # unused for flagging

    def flag_within(self, tag=None, **kwargs):
        """
        Flag a value within a plus or minus for a given key(s).
        e.g. `key1 = [value1,epsilon1], key2 = [value2,epsilon2], ...`
        Will select data
        `value1-epsilon1 <= data1 <= value1+epsilon1,`
        `value2-epsilon2 <= data2 <= value2+epsilon2,...`

        Parameters
        ----------
        tag : str, optional
            An identifying tag by which the rule may be referred to later.
            If None, a  randomly generated tag will be created.
        key : str
            The key (SDFITS column name or other supported key)
        value : array-like
            Tuple or list giving the value and epsilon

        Returns
        -------
        None.

        """
        self._base_select_within(tag, **kwargs)
        idx = len(self._table) - 1
        self._flag_channel_selection[idx] = ALL_CHANNELS
        self._selection_rules[idx]["CHAN"] = ALL_CHANNELS
        self._channel_selection = None  # unused for flagging

    def read(self, fileobj, **kwargs):
        """Read a GBTIDL flag file and instantiate Flag object.

        Parameters
        ----------
        fileobj : str, file-like or `pathlib.Path`
            File to read.  If a file object, must be opened in a
            readable mode.
        **kwargs : dict
            Extra keyword arguments to apply to the flag rule.  (This is mainly for internal use.)

        Returns
        -------
        None.

        """
        # GBTIDL flag files two sections [header] and [flags]
        # In the [header] section is information about file creation.
        # The [flags] section containes the flag table
        # The table has 10 columns. Its rows have vertical bar (|) separated columns, while
        # the table header is separated by commas and begins with a #
        # The columns are:
        #
        # ID - flag ID number, same as dysh's flag rule `id`
        # RECNUM - range of the selected record numbers given as low:high inclusive
        # SCAN - range of the selected scan numbers given as low:high inclusive
        # INTNUM - range of the selected integration numbers given as low:high inclusive
        # PLNUM - range of the selected polarization numbers given as low:high inclusive
        # IFNUM - range of the selected IF numbers given as low:high inclusive
        # BCHAN - beginning channel flagged (inclusive, starting from zero)
        # ECHAN - end channel flagged (inclusive)
        # IDSTRING - Reason for flagging, same as dysh's flag rule `tag`
        #
        # Numeric alues can be a single integer or comma-separated list of integers.  If BCHAN and ECHAN
        # are a comma-separated list then they must be pair up as [bchan_i,echan+i]
        # A wildcard appears in a column if it had no selection (meaning all values were selected).
        # Example file:
        # [header]
        # created = Wed Jan  5 16:48:37 2022
        # version = 1.0
        # created_by = sdfits
        # [flags]
        # #RECNUM,SCAN,INTNUM,PLNUM,IFNUM,FDNUM,BCHAN,ECHAN,IDSTRING
        # *|6|*|*|2|0|3072|3072|VEGAS_SPUR
        #
        # It is possible there is a space after the *GBTIDL flag files can also indicate ranges with a : and can indicate upper or lower limits
        # by not including a number. For instancer here is scan range 42 to 51 and channel range with
        # lower limit of 2299
        # *|20|42:51|*|*|*|2299|*|unspecified

        # Because the table header and table row delimeters are different,
        # Table.read() can't work.  So construct it row by row.
        f = open(fileobj, mode="r")
        lines = f.read().splitlines()  # gets rid of \n
        f.close()
        header = ["RECNUM", "SCAN", "INTNUM", "PLNUM", "IFNUM", "FDNUM", "BCHAN", "ECHAN", "IDSTRING"]
        for l in lines[lines.index("[flags]") + 1 :]:
            vdict = {}
            if l.startswith("#"):
                # its the header
                colnames = l[1:].split(",")
                if colnames != header:
                    raise Exception(f"Column names {colnames} do not match expectated {header}")
            else:
                values = l.split("|")
                for i, v in enumerate(values):
                    if v.strip() == "*":
                        continue
                    else:
                        if header[i] == "IDSTRING":
                            vdict[header[i]] = v
                        else:
                            # handle comma-separated lists
                            if "," in v:
                                vdict[header[i]] = [int(float(x)) for x in v.split(",")]
                            # handle colon-separated ranges by expanding into a comma-separated list.
                            elif ":" in v:
                                vdict[header[i]] = [int(float(x)) for x in range(*map(int, v.split(":")))] + [
                                    int(v.split(":")[-1])
                                ]
                            # handle single values
                            else:
                                vdict[header[i]] = int(float(v))

                # our tag is gbtidl's idstring
                tag = vdict.pop("IDSTRING", None)
                bchan = vdict.pop("BCHAN", None)
                echan = vdict.pop("ECHAN", None)
                if bchan is not None and echan is not None:
                    if not isinstance(bchan, list):
                        bchan = [bchan]
                    bchan = [int(float(x)) for x in bchan]
                    if not isinstance(echan, list):
                        echan = [echan]
                    echan = [int(float(x)) for x in echan]
                    # pair up echan and bchan
                    vdict["channel"] = list(zip(bchan, echan))
                elif bchan is not None and echan is None:
                    if not isinstance(bchan, list):
                        bchan = [bchan]
                    bchan = [int(float(x)) for x in bchan]
                    echan = [2**25] * len(
                        bchan
                    )  # Set to a large number so it effectively spans the whole range from `bchan`.
                    vdict["channel"] = tuple(zip(bchan, echan))
                elif bchan is None and echan is not None:
                    if not isinstance(echan, list):
                        echan = [echan]
                    echan = [int(float(x)) for x in echan]
                    bchan = [0] * len(echan)
                    vdict["channel"] = tuple(zip(bchan, echan))

                if kwargs is not None:
                    vdict.update(kwargs)
                logger.debug(f"flag({tag=},{vdict})")
                self.flag(tag=tag, **vdict)



import warnings
from unittest.mock import patch

import astropy.units as u
import numpy as np
import pytest
from astropy.io import fits

from dysh.fits.gbtfitsload import GBTFITSLoad
from dysh.spectra.spectrum import IGNORE_ON_COPY, Spectrum, average_spectra
from dysh.util import get_project_testdata


def fit_gauss(spectrum):
    """
    Fit a Gaussian.
    """

    from astropy.modeling import models
    from specutils.fitting import fit_lines

    g_init = models.Gaussian1D(
        amplitude=spectrum.flux.max(), mean=spectrum.spectral_axis.mean(), stddev=spectrum.meta["FREQRES"] * u.Hz
    )
    g_fit = fit_lines(spectrum, g_init)

    return g_fit


def compare_spectrum(one, other, ignore_history=False, ignore_comments=False):
    """ """

    for k, v in vars(one).items():
        if k in IGNORE_ON_COPY:
            continue
        if ignore_history and k == "_history":
            continue
        if ignore_history and k == "_comments":
            continue
        elif k in ["_wcs"]:
            v.to_header() == vars(other)[k].to_header()
        elif k in ["_spectral_axis"]:
            for k_, v_ in vars(v).items():
                assert v_ == vars(vars(other)[k])[k_]
        else:
            assert v == vars(other)[k]


def loadfits(fname):
    hdu = fits.open(fname)
    spec = hdu[1].data["DATA"][0]
    hdu.close()
    return spec


class TestSpectrum:
    def setup_method(self):
        data_dir = get_project_testdata() / "AGBT05B_047_01"
        sdf_file = data_dir / "AGBT05B_047_01.raw.acs"
        sdf = GBTFITSLoad(sdf_file)
        self.getps0 = sdf.getps(scan=51, plnum=0)
        self.ps0 = self.getps0.timeaverage()
        self.getps1 = sdf.getps(scan=51, plnum=1)
        self.ps1 = self.getps1.timeaverage()
        self.ss = self.ps0._copy()  # Synthetic one.
        x = np.arange(0, len(self.ss.data))
        fwhm = 5
        stdd = fwhm / 2.35482
        mean = int(x.mean())
        self.ss._data = 1 * np.exp(-0.5 * (x - mean) ** 2 / stdd**2)
        self.ss.meta["FREQRES"] = abs(self.ss.meta["CDELT1"])
        self.ss.meta["FWHM"] = fwhm
        self.ss.meta["CENTER"] = self.ss.spectral_axis[mean].value
        self.ss.meta["STDD"] = stdd

    def test_add(self):
        """Test that we can add two `Spectrum`."""
        addition = self.ps0 + self.ps1

        assert addition.meta["EXPOSURE"] == (self.ps0.meta["EXPOSURE"] + self.ps1.meta["EXPOSURE"])
        assert np.all(addition.flux.value == (self.ps0.flux.value + self.ps1.flux.value))
        assert addition.flux.unit == self.ps0.flux.unit
        assert addition.velocity_frame == self.ps0.velocity_frame
        compare_spectrum(self.ps0, addition)

    def test_add_scalar(self):
        """Test that we can add a scalar to a `Spectrum`."""
        addition = self.ps0 + 10.0

        assert addition.meta["EXPOSURE"] == self.ps0.meta["EXPOSURE"]
        assert np.all(addition.flux.value == (self.ps0.flux.value + 10.0))
        assert addition.flux.unit == self.ps0.flux.unit
        assert addition.velocity_frame == self.ps0.velocity_frame
        compare_spectrum(self.ps0, addition)

    def test_radd_scalar(self):
        """Test that we can add a scalar to a `Spectrum`."""
        addition = 10 + self.ps0

        assert addition.meta["EXPOSURE"] == self.ps0.meta["EXPOSURE"]
        assert np.all(addition.flux.value == (self.ps0.flux.value + 10.0))
        assert addition.flux.unit == self.ps0.flux.unit
        assert addition.velocity_frame == self.ps0.velocity_frame
        compare_spectrum(self.ps0, addition)

    def test_sub(self):
        """Test that we can subtract two `Spectrum`."""
        subtraction = self.ps0 - self.ps1

        assert subtraction.meta["EXPOSURE"] == (self.ps0.meta["EXPOSURE"] + self.ps1.meta["EXPOSURE"])
        assert np.all(subtraction.flux.value == (self.ps0.flux.value - self.ps1.flux.value))
        assert subtraction.flux.unit == self.ps0.flux.unit
        assert subtraction.velocity_frame == self.ps0.velocity_frame
        compare_spectrum(self.ps0, subtraction)

    def test_sub_scalar(self):
        """Test that we can subtract a scalar from a `Spectrum`."""
        subtraction = self.ps0 - 10.0

        assert subtraction.meta["EXPOSURE"] == self.ps0.meta["EXPOSURE"]
        assert np.all(subtraction.flux.value == (self.ps0.flux.value - 10.0))
        assert subtraction.flux.unit == self.ps0.flux.unit
        assert subtraction.velocity_frame == self.ps0.velocity_frame
        compare_spectrum(self.ps0, subtraction)

    def test_rsub_scalar(self):
        """Test that we can subtract a scalar from a `Spectrum`."""
        subtraction = 10.0 - self.ps0

        assert subtraction.meta["EXPOSURE"] == self.ps0.meta["EXPOSURE"]
        assert np.all(subtraction.flux.value == (10.0 - self.ps0.flux.value))
        assert subtraction.flux.unit == self.ps0.flux.unit
        assert subtraction.velocity_frame == self.ps0.velocity_frame
        compare_spectrum(self.ps0, subtraction)

    def test_mul(self):
        """Test that we can multiply two `Spectrum`."""
        multiplication = self.ps0 * self.ps1

        assert np.all(multiplication.flux.value == (self.ps0.flux.value * self.ps1.flux.value))
        assert multiplication.flux.unit == self.ps0.flux.unit * self.ps1.flux.unit
        assert multiplication.velocity_frame == self.ps0.velocity_frame
        compare_spectrum(self.ps0, multiplication)

    def test_mul_scalar(self):
        """Test that we can multiply a `Spectrum` and a scalar."""
        multiplication = self.ps0 * 1.0

        assert np.all(multiplication.flux.value == (self.ps0.flux.value))
        assert multiplication.flux.unit == self.ps0.flux.unit
        assert multiplication.velocity_frame == self.ps0.velocity_frame
        compare_spectrum(self.ps0, multiplication)

    def test_rmul_scalar(self):
        """Test that we can multiply a `Spectrum` and a scalar."""
        multiplication = 1.0 * self.ps0

        assert np.all(multiplication.flux.value == (self.ps0.flux.value))
        assert multiplication.flux.unit == self.ps0.flux.unit
        assert multiplication.velocity_frame == self.ps0.velocity_frame
        compare_spectrum(self.ps0, multiplication)

    def test_div(self):
        """Test that we can divide two `Spectrum`."""
        division = self.ps0 / self.ps1

        assert np.all(division.flux.value == (self.ps0.flux.value / self.ps1.flux.value))
        assert division.flux.unit == self.ps0.flux.unit / self.ps1.flux.unit
        compare_spectrum(self.ps0, division)

    def test_div_scalar(self):
        """Test that we can divide a `Spectrum` by a scalar."""
        division = self.ps0 / 1.0

        assert np.all(division.flux.value == (self.ps0.flux.value))
        assert division.flux.unit == self.ps0.flux.unit
        assert division.velocity_frame == self.ps0.velocity_frame
        compare_spectrum(self.ps0, division)

    def test_write_read_fits(self, tmp_path):
        """Test that we can read fits files written by dysh"""
        s = self.ps1
        o = tmp_path / "sub"
        o.mkdir()
        file = o / "test_spectrum_write.fits"
        s.write(file, format="fits", overwrite=True)
        s2 = Spectrum.read(file, format="fits")
        assert np.all(s.data == s2.data)
        assert s.target == s2.target
        assert np.all(s2.spectral_axis == s.spectral_axis)
        # This test will generally fail because SITELONG, SITELAT, SITEELEV
        # don't have enough precision to match exactly our known GBT coordinates.
        # @todo make a close_enough comparison by differencing the observer
        # attributes
        # if s2.observer is not None:
        #    assert s.observer == s2.observer

    def test_write_read_ascii(self, tmp_path):
        fmt = [
            "basic",
            "ascii.commented_header",
            "commented_header",
            "ascii.fixed_width",
            "ascii.ipac",
            "ipac",
            "votable",
            "ecsv",
            "mrt",
        ]
        s = self.ps1
        o = tmp_path / "sub"
        o.mkdir()
        for f in fmt:
            file = o / f"testwrite.{f}"
            s.write(file, format=f, overwrite=True)
            # ECSV is the only ascii format that can
            # complete a roundtrip unscathed.
            # (See https://docs.astropy.org/en/latest/io/unified.html#table-io)
            if f == "ecsv":
                s2 = Spectrum.read(file, format=f)
                assert np.all(s.data == s2.data)
                assert np.all(s.spectral_axis == s2.spectral_axis)
                assert s.target == s2.target
        # Test reading in a GBTIDL ascii file
        gbtidl_file = get_project_testdata() / "gbtidl_spectra/onoff-L_gettp_156_intnum_0_LSR.ascii"
        s2 = Spectrum.read(gbtidl_file, format="gbtidl")
        assert s2.meta["SCAN"] == 156
        assert s2.meta["OBJECT"] == "NGC2782"
        # veldef can't be determined from header.
        assert s2.flux.unit == u.ct
        assert s2.flux[0].value == 3608710.0
        assert s2.spectral_axis.unit == u.GHz
        # Now try a gbtidl file that is gzipped and Ta units
        gbtidl_file = get_project_testdata() / "gbtidl_spectra/onoff-L_getps_152_OPTI-HEL.ascii.gz"
        s2 = Spectrum.read(gbtidl_file, format="gbtidl")
        assert s2.meta["SCAN"] == 152
        assert s2.meta["OBJECT"] == "NGC2415"
        assert s2.meta["VELDEF"] == "OPTI-HEL"
        assert s2.flux.unit == u.K
        assert s2.flux[0].value == -0.1042543
        assert s2.spectral_axis.unit == u.Unit("km/s")

    def test_history_and_comments(self):
        s = self.ps1
        s.baseline(2, remove=True)
        print(s.comments)
        s.add_comment("I removed a baseline")
        # This tests that the baseline command self-logged to history
        # AND that order is preserved because
        # the baseline history should be the last entry
        # print(s.history)
        assert "baseline(2,remove=True,)" in s.history[-1]
        print(s.comments)
        assert "I removed a baseline" in s.comments

    @patch("dysh.plot.specplot.plt.show")
    def test_slice(self, mock_show, tmp_path):
        """
        Test that we can slice a `Spectrum` using channels or units.
        For units we only consider frequencies for now.
        """
        meta_ignore = ["CRPIX1", "CRVAL1"]
        spec_pars = ["_target", "_velocity_frame", "_observer", "_obstime"]
        s = slice(1000, 1100, 1)
        trimmed = self.ps0[s]
        assert trimmed.flux[0] == self.ps0.flux[s.start]
        assert trimmed.flux[-1] == self.ps0.flux[s.stop - 1]
        assert np.all(trimmed.flux == self.ps0.flux[s])
        # The slicing changes the values at the micro Hz level.
        assert np.all(trimmed.spectral_axis.value - self.ps0.spectral_axis[s].value < 1e-5)
        # Check meta values. The trimmed spectrum has an additional
        # key: 'original_wcs'.
        for k, v in self.ps0.meta.items():
            if k not in meta_ignore:
                assert trimmed.meta[k] == v
        # Check additional object properties.
        # Not all of them make sense, since their shapes will be different.
        for k in spec_pars:
            assert vars(trimmed)[k] == vars(self.ps0)[k]
        # Check that we can plot.
        trimmed.plot(xaxis_unit="km/s", yaxis_unit="mK", vel_frame="itrs")
        # Check that we can write.
        o = tmp_path / "sub"
        o.mkdir()
        out = o / "test_spec_slice_write.fits"
        trimmed.write(out, format="fits", overwrite=True)
        # Check that we can read it back.
        trimmed_read = Spectrum.read(out, format="fits")
        assert np.all(trimmed.flux == trimmed_read.flux)
        assert np.all(trimmed.spectral_axis == trimmed_read.spectral_axis)
        assert trimmed.target == trimmed_read.target

        # Now slice using units.
        # Hz.
        spec_ax = self.ps0.spectral_axis
        trimmed_nu = self.ps0[spec_ax[s.start].to("Hz") : spec_ax[s.stop].to("Hz")]
        assert np.all(trimmed_nu.flux == self.ps0.flux[s])
        assert np.all(trimmed_nu.spectral_axis.value - self.ps0.spectral_axis[s].value < 1e-5)
        for k, v in self.ps0.meta.items():
            if k not in meta_ignore:
                assert trimmed_nu.meta[k] == v
        for k in spec_pars:
            assert vars(trimmed_nu)[k] == vars(self.ps0)[k]
        trimmed_nu.plot(xaxis_unit="km/s", yaxis_unit="mK")

        # km/s.
        spec_ax = self.ps0.spectral_axis.to("km/s")
        trimmed_vel = self.ps0[spec_ax[s.start] : spec_ax[s.stop]]
        assert np.all(trimmed_vel.flux == self.ps0.flux[s])
        assert np.all(trimmed_vel.spectral_axis.value - self.ps0.spectral_axis[s].value < 1e-5)
        for k, v in self.ps0.meta.items():
            if k not in meta_ignore:
                assert trimmed_vel.meta[k] == v
        for k in spec_pars:
            assert vars(trimmed_vel)[k] == vars(self.ps0)[k]
        trimmed_vel.plot(xaxis_unit="MHz", yaxis_unit="mK")

        # m.
        spec_ax = self.ps0.spectral_axis.to("m")
        trimmed_wav = self.ps0[spec_ax[s.start] : spec_ax[s.stop]]
        assert np.all(trimmed_wav.flux == self.ps0.flux[s])
        assert np.all(trimmed_wav.spectral_axis.value - self.ps0.spectral_axis[s].value < 1e-5)

    def test_smooth(self):
        """Test for smooth with `decimate=0`"""
        width = 10
        ss = self.ps0.smooth("gauss", width)
        assert ss.meta["CDELT1"] == self.ps0.meta["CDELT1"] * width
        assert ss.meta["FREQRES"] == pytest.approx(abs(self.ps0.meta["CDELT1"]) * width)
        assert np.diff(ss.spectral_axis).mean().value == ss.meta["CDELT1"]
        assert ss._resolution == pytest.approx(1)
        assert ss.velocity_frame == self.ps0.velocity_frame
        assert ss.doppler_convention == self.ps0.doppler_convention
        assert ss.observer.frame_attributes == self.ps0.observer.frame_attributes

        # Now, change the reference frame and see if it still works.
        from dysh.coordinates import astropy_frame_dict

        s = Spectrum.fake_spectrum()
        for frame in astropy_frame_dict.keys():
            try:
                s.set_frame(frame)
            except Exception:
                print(f"set_frame fails for: {frame}")
                continue
            print(f"frame set to: {frame}")
            s.smooth("box", 3)

    def test_smooth_decimate(self):
        """Test for smooth with `decimate!=width`."""
        width = 10
        decimate = 8
        ss = self.ps0.smooth("gauss", width, decimate)
        assert ss.meta["CDELT1"] == self.ps0.meta["CDELT1"] * decimate
        assert ss.meta["FREQRES"] == pytest.approx(abs(self.ps0.meta["CDELT1"]) * width)
        assert np.diff(ss.spectral_axis).mean().value == ss.meta["CDELT1"]
        assert ss._resolution == pytest.approx(width / decimate)

        # Now with synthetic data.
        sss = self.ss.smooth("gauss", width, decimate)
        assert sss.meta["CDELT1"] == self.ss.meta["CDELT1"] * decimate
        assert sss.meta["FREQRES"] == pytest.approx(abs(self.ss.meta["CDELT1"]) * width, abs=100)
        assert np.diff(sss.spectral_axis).mean().value == sss.meta["CDELT1"]
        assert sss._resolution == pytest.approx(width / decimate, abs=1e-2)
        # Also check the line properties.
        g_fit = fit_gauss(sss)
        fwhm = g_fit.stddev.value * 2.35482
        assert g_fit.mean.value == pytest.approx(self.ss.meta["CENTER"])
        assert np.sqrt(fwhm**2 - sss.meta["FREQRES"] ** 2) == pytest.approx(
            abs(self.ss.meta["CDELT1"]) * self.ss.meta["FWHM"], abs=abs(self.ss.meta["CDELT1"]) / 9.0
        )
        assert ss.velocity_frame == self.ps0.velocity_frame
        assert ss.doppler_convention == self.ps0.doppler_convention
        assert ss.observer.frame_attributes == self.ps0.observer.frame_attributes

    def test_smooth_nodecimate(self):
        """Test for smooth without decimation."""
        width = 10
        decimate = -1
        ss = self.ps0.smooth("gauss", width, decimate)
        assert ss.meta["CDELT1"] == self.ps0.meta["CDELT1"]
        assert ss.meta["FREQRES"] == pytest.approx(abs(self.ps0.meta["CDELT1"]) * width)
        assert np.diff(ss.spectral_axis).mean().value == ss.meta["CDELT1"]
        assert ss._resolution == pytest.approx(width / abs(decimate))

    def test_smooth_multi(self):
        """Test for multiple passes of smooth."""
        widths = [10, 15, 15.1]
        decimate = -1

        # Check fitter first.
        g_fit = fit_gauss(self.ss)
        assert g_fit.stddev.value * 2.35482 == pytest.approx(
            abs(self.ss.meta["CDELT1"]) * self.ss.meta["FWHM"], g_fit.stddev.std * 2.35482
        )

        # Now smooth the same Spectrum multiple times.
        sss = self.ss._copy()
        for w in widths:
            sss = sss.smooth("gauss", w, decimate=decimate)
            g_fit = fit_gauss(sss)
            fwhm = g_fit.stddev.value * 2.35482
            assert sss.meta["FREQRES"] == pytest.approx(abs(self.ss.meta["CDELT1"]) * w)
            assert np.sqrt(fwhm**2 - sss.meta["FREQRES"] ** 2) == pytest.approx(
                abs(self.ss.meta["CDELT1"]) * self.ss.meta["FWHM"], abs=abs(self.ss.meta["CDELT1"]) / 9.0
            )
            assert g_fit.mean.value == pytest.approx(self.ss.meta["CENTER"])

    def test_smooth_and_slice(self):
        """Test for slicing after smoothing."""
        width = 10
        decimate = 8
        s = slice(-500 * u.km / u.s, 500 * u.km / u.s)
        sss = self.ss.smooth("gauss", width, decimate)
        ssss = sss[s]
        g_fit = fit_gauss(sss)
        fwhm = g_fit.stddev.value * 2.35482
        assert ssss.meta["CDELT1"] == self.ss.meta["CDELT1"] * decimate
        assert ssss.meta["FREQRES"] == pytest.approx(abs(self.ss.meta["CDELT1"]) * width, abs=100)
        assert np.diff(sss.spectral_axis).mean().value == sss.meta["CDELT1"]
        assert sss._resolution == pytest.approx(width / decimate, abs=1e-2)
        assert g_fit.mean.value == pytest.approx(self.ss.meta["CENTER"])
        assert np.sqrt(fwhm**2 - sss.meta["FREQRES"] ** 2) == pytest.approx(
            abs(self.ss.meta["CDELT1"]) * self.ss.meta["FWHM"], abs=abs(self.ss.meta["CDELT1"]) / 9.0
        )

    def test_shift(self):
        """Test the shift method against the results produced by GBTIDL"""

        # Prepare test data.
        filename = get_project_testdata() / "TGBT21A_501_11/TGBT21A_501_11_ifnum_0_int_0-2_getps_152_plnum_0.fits"
        sdf = GBTFITSLoad(filename)
        nchan = sdf["DATA"].shape[-1]
        spec = Spectrum.fake_spectrum(nchan=nchan, seed=1)
        spec.data[nchan // 2 - 5 : nchan // 2 + 6] = 10
        org_spec = spec._copy()
        # The next two lines were used to create the input for GBTIDL.
        # sdf["DATA"] = [spec.data]
        # sdf.write("shift_testdata.fits")

        # Apply method to be tested.
        shift = 5.5
        spec = spec.shift(shift)

        # Internal tests.
        assert np.all(np.isnan(spec[: int(np.round(shift))].data))

        # Load GBTIDL answer.
        with fits.open(get_project_testdata() / "gshift_box.fits") as hdu:
            table = hdu[1].data
        gbtidl = table["DATA"][0]

        # Compare.
        # Ignore the edge channels to avoid edge effects.
        diff = (spec.data - gbtidl)[10:-10]
        assert np.all(abs(diff) < 5e-4)

        assert spec.meta["CRPIX1"] == org_spec.meta["CRPIX1"] + shift
        assert spec.spectral_axis[0].to("Hz").value == (
            org_spec.spectral_axis[0].to("Hz").value - spec.meta["CDELT1"] * shift
        )

    def test_find_shift(self):
        """
        Test the find_shift method.
        * Test that the shift with respect to itself is zero.
        * Test that it can find an integer shift.
        * Test that it can find a fractional shift.
        * Test that it can find a shift in velocity units.
        * Test that it can find a shift in a different frame.
        """
        spec = Spectrum.fake_spectrum(seed=1)

        # Shift should be zero.
        assert spec.find_shift(spec) == pytest.approx(0)

        chan_wid = np.mean(np.diff(spec._spectral_axis))

        # Shift by one channel.
        spec2 = spec._copy()
        spec2._spectral_axis = spec2.spectral_axis.replicate(value=spec2.spectral_axis + chan_wid)
        assert spec.find_shift(spec2) == pytest.approx(-1)

        # Shift by one and a half channels.
        spec3 = spec._copy()
        spec3._spectral_axis = spec3.spectral_axis.replicate(value=spec3.spectral_axis + chan_wid * 1.5)
        assert spec.find_shift(spec3) == pytest.approx(-1.5)

        # Shift in velocity.
        spec4 = spec._copy()
        velo = spec4.spectral_axis.replicate(value=spec4.spectral_axis.to("km/s"))
        dvel = velo[1] - velo[0]
        spec4._spectral_axis = spec4.spectral_axis.replicate(value=velo + dvel * 0.5)
        assert spec.find_shift(spec4) == pytest.approx(-0.5)

        # Shift in a different frame.
        assert spec.find_shift(spec4, frame="lsrk") == pytest.approx(-0.5)

    def test_align_to(self):
        """
        Tests for align_to method.
        * Test that align_to itself does not change the spectrum.
        * Test that aligning to a spectrum with a integer shift preserves the data.
        * Test that aligning to a spectrum with a fractional shift preserves signal amplitude within errorbars.
        """

        spec = Spectrum.fake_spectrum(nchan=1024, seed=1)
        org_spec = spec._copy()

        # Align to itself.
        spec = spec.align_to(spec)
        compare_spectrum(spec, org_spec, ignore_history=True)
        assert np.all((spec - org_spec).data == 0)

        # Align to a shifted version.
        shift = 5
        spec = spec.shift(shift)
        assert np.all((spec.data[shift:] - org_spec.data[:-shift]) == 0.0)

        # Align to a shifted version with signal.
        fshift = 0.5
        spec = self.ss._copy()
        org_spec = spec._copy()
        spec = spec.shift(shift + fshift)
        # The amplitude of the signal will decrease because of the sampling.
        tol = np.sqrt(
            (1 - np.exp(-0.5 * (fshift) ** 2 / spec.meta["STDD"] ** 2)) ** 2.0
            + (np.nanstd(spec.data[: len(spec.data) - 50])) ** 2.0
        )
        assert spec.max().value == pytest.approx(org_spec.max().value, abs=3 * tol)

    def test_average_spectra(self):
        """
        Tests for average_spectra.
        Although not a class method of `Spectra` it is included here to reuse the setup method of the test.
        * Test that it does not crash.
        * Test that it does not crash whit alignment.
        """

        ps0_org = self.ps0._copy()
        ps1_org = self.ps1._copy()

        avg = average_spectra((self.ps0, self.ps1))
        avg2 = self.ps0.average((self.ps1))
        compare_spectrum(avg, avg2, ignore_history=True, ignore_comments=True)

        avg = average_spectra((self.ps0, self.ps1), align=True)
        compare_spectrum(ps0_org, self.ps0, ignore_history=True, ignore_comments=True)
        compare_spectrum(ps1_org, self.ps1, ignore_history=True, ignore_comments=True)

    def test_spectrum_with_frame(self):
        """Regression test for issue #401 to ensure Spectrum.with_frame functions as advertised.
        https://github.com/GreenBankObservatory/dysh/issues/401
        """
        spec = Spectrum.fake_spectrum()
        # Ensure that repeated changes of frame to the same frame do note
        # change after the first transform
        s1 = spec.with_frame("lsrk")
        s2 = s1.with_frame("lsrk")
        assert all(s1.spectral_axis == s2.spectral_axis)
        assert s2.velocity_frame == "lsrk"
        assert s2.meta["VELDEF"][4:] == "-LSR"

        # Test that topographic results in an Exception because
        # users must provide an ITRS coordinate instance in that case.
        # First change the spectrum frame to somthing else because
        # if it is already topo/itrs, then the error is circumvented
        spec2 = Spectrum.fake_spectrum()
        spec2.set_frame("gcrs")
        with pytest.raises(ValueError):
            spec2.set_frame("topo")

        # Setting a new frame to the old frame does NOT result in an
        # identical observer attribute on the resultant SpectralAxis.
        # See https://github.com/astropy/astropy/issues/17506
        # This test ensures that the difference remains small.

        location_diff = np.sqrt(
            (s1.observer.x - s2.observer.x) ** 2
            + (s1.observer.y - s2.observer.y) ** 2
            + (s1.observer.z - s2.observer.z) ** 2
        )
        velocity_diff = np.sqrt(
            (s1.observer.v_x - s2.observer.v_x) ** 2
            + (s1.observer.v_y - s2.observer.v_y) ** 2
            + (s1.observer.v_z - s2.observer.v_z) ** 2
        )
        assert location_diff < 1.0e-5 * u.m
        assert velocity_diff < 2e-8 * u.km / u.s

    def test_velocity_axis_to(self):
        """Regression test for issue 372 https://github.com/GreenBankObservatory/dysh/issues/372
        Calling velocity_axis_to should not change the object spectral axis.
        """
        spec = Spectrum.fake_spectrum()
        id1 = id(spec.spectral_axis)
        sa = spec.velocity_axis_to(toframe="lsrk")
        assert id(spec.spectral_axis) == id1
        assert id(sa) != id1
        # converting to identical frame should not change the values in the spectral axis
        sa = spec.velocity_axis_to(toframe="topo")
        assert all(sa == spec.spectral_axis)

    def test_baseline(self):
        """Test for comparing GBTIDL baseline to Dysh baselines"""

        def test_single_baseline(sdf, ex_reg, order, model, gbtidl_bmodel, exclude_region_upper_bounds=True):
            """For use with TestSpectrum.test_baseline()"""
            dysh_spec = sdf.getspec(0)
            temp_bmodel = np.copy(dysh_spec.data)
            dysh_spec.baseline(
                order, ex_reg, remove=True, model=model, exclude_region_upper_bounds=exclude_region_upper_bounds
            )
            dysh_bmodel = temp_bmodel - np.copy(dysh_spec.data)
            diff = np.sum(np.abs(dysh_bmodel - gbtidl_bmodel))
            assert diff < 1.5e-6

        data_dir = get_project_testdata() / "AGBT17A_404_01"
        sdf_file = data_dir / "AGBT17A_404_01_scan_19_prebaseline.fits"
        sdf = GBTFITSLoad(sdf_file)
        gbtidl_two_reg = loadfits(data_dir / "AGBT17A_404_01_scan_19_bmodel.fits")
        gbtidl_no_reg = loadfits(data_dir / "AGBT17A_404_01_scan_19_noregion_bmodel.fits")

        order = 3
        ex_reg = [(0, 99), (381, 449), (721, 819)]

        test_single_baseline(sdf, ex_reg, order, "chebyshev", gbtidl_two_reg)
        test_single_baseline(sdf, ex_reg, order, "legendre", gbtidl_two_reg)
        test_single_baseline(sdf, ex_reg, order, "hermite", gbtidl_two_reg)
        test_single_baseline(sdf, ex_reg, order, "polynomial", gbtidl_two_reg)

        ex_reg = None
        test_single_baseline(sdf, ex_reg, order, "chebyshev", gbtidl_no_reg)

        # Test that no warnings are issued.
        with warnings.catch_warnings():
            warnings.simplefilter("error")
            self.ps0.baseline(model="poly", degree=2)

        # Test baseline removal with velocity units and known anwser.
        s = Spectrum.fake_spectrum()
        s.spectral_axis.to("km/s")
        s._spectral_axis = s.spectral_axis.to("km/s")
        # Add a polynomial.
        pcoef = np.array([1, 5, 10])
        s += np.poly1d(pcoef)(np.linspace(1, -1, len(s.spectral_axis)))
        s.baseline(degree=2, model="poly", remove=True)
        for pn in s._baseline_model.unitless_model.param_names:
            fit_val = getattr(s._baseline_model.unitless_model, pn).value
            in_val = pcoef[::-1][int(pn[-1])]
            assert (in_val - fit_val) / in_val < 0.1

        # Test with units. Nothing to compare to though.
        dysh_spec = sdf.getspec(0)
        kms = u.km / u.s
        ex_reg = [(0 * kms, 4200 * kms), (6000 * kms, 7000 * kms), (8800 * kms, 90000 * kms)]
        dysh_spec.baseline(order, ex_reg, model="chebyshev")
        ex_reg = [(1 * u.GHz, 1.38 * u.GHz), (1.388 * u.GHz, 1.392 * u.GHz), (1.4 * u.GHz, 2 * u.GHz)]
        dysh_spec.baseline(order, ex_reg, model="chebyshev")
        ex_reg = [21 * u.cm, 21.5 * u.cm]
        dysh_spec.baseline(order, ex_reg, model="chebyshev")

    def test_baseline_include(self):
        """
        Test for comparing GBTIDL baseline to dysh baselines
        using inclusion regions.
        """

        def test_single_baseline(sdf, in_reg, order, model, gbtidl_bmodel, exclude_region_upper_bounds=True):
            """For use with TestSpectrum.test_baseline()"""
            dysh_spec = sdf.getspec(0)
            temp_bmodel = np.copy(dysh_spec.data)
            dysh_spec.baseline(
                order, include=in_reg, remove=True, model=model, exclude_region_upper_bounds=exclude_region_upper_bounds
            )
            dysh_bmodel = temp_bmodel - np.copy(dysh_spec.data)
            diff = np.sum(np.abs(dysh_bmodel - gbtidl_bmodel))
            assert diff < 1.5e-6

        data_dir = get_project_testdata() / "AGBT17A_404_01"
        sdf_file = data_dir / "AGBT17A_404_01_scan_19_prebaseline.fits"
        sdf = GBTFITSLoad(sdf_file)
        gbtidl_two_reg = loadfits(data_dir / "AGBT17A_404_01_scan_19_bmodel.fits")
        gbtidl_no_reg = loadfits(data_dir / "AGBT17A_404_01_scan_19_noregion_bmodel.fits")

        order = 3
        in_reg = [(99, 381), (449, 721)]

        test_single_baseline(sdf, in_reg, order, "chebyshev", gbtidl_two_reg)
        test_single_baseline(sdf, in_reg, order, "legendre", gbtidl_two_reg)
        test_single_baseline(sdf, in_reg, order, "hermite", gbtidl_two_reg)

        # Test with units. Nothing to compare to though.
        dysh_spec = sdf.getspec(0)
        kms = u.km / u.s
        in_reg = [(4200 * kms, 6000 * kms), (7000 * kms, 8800 * kms)]
        dysh_spec.baseline(order, include=in_reg, model="chebyshev")
        in_reg = [(1.38 * u.GHz, 1.388 * u.GHz), (1.392 * u.GHz, 1.4 * u.GHz)]
        dysh_spec.baseline(order, include=in_reg, model="chebyshev")
        in_reg = [21 * u.cm, 21.5 * u.cm]
        dysh_spec.baseline(order, include=in_reg, model="chebyshev")



import argparse
import subprocess


def parse_args():
    parser = argparse.ArgumentParser(
        description=(
            "dysh in Jupyter Lab.\n\nAll CLI arguments other than those defined below are passed through "
            "to jupyter; see $ jupyter lab --help for more details"
        )
    )

    return parser.parse_known_args()


def main():
    args, remaining_args = parse_args()

    subprocess.run(["jupyter", "lab", *remaining_args])


if __name__ == "__main__":
    main()



from pathlib import Path

import numpy as np

import dysh.util.files as duf


class TestUtil:
    """Test dysh.util files functions"""

    def test_dysh_data(self):
        """Test dysh_data   (only limited testing possible)"""
        assert duf.dysh_data() == None
        # test=
        f1 = duf.dysh_data(test="getps")
        assert f1.exists() == True
        # example=
        #   given the dynamic nature, not sure if we should test example="getps"
        #   since it needs either $DYSH_DATA or /home/dysh/example_data
        # acccept=
        #   skipping
        # sdfits=
        #   skipping
        # dysh_data=
        assert duf.dysh_data("foo.fits", dysh_data="/tmp") == None
        #   this assume DYSH_DATA is not present
        f2 = duf.dysh_data(example="test1")
        assert f2 == Path("AGBT05B_047_01.raw.acs.fits")
